#include "Scene.h"


Scene::Scene(const char * descName) {
  init();
  
  SceneDescription  desc;
  
  std::ifstream ifs(descName);
  boost::archive::text_iarchive ia(ifs);
  ia >> desc;

  tetmesh = TetMesh::ptr(new TetMesh(desc.tetMesh.mat));
  tetmesh->loadobj(desc.tetMesh.fileName.c_str());
  tetmesh->fracThresh = desc.tetMesh.fracThresh;
    
  if (desc.tetBoundMesh.fileName.size()) {
    boundMesh.loadobj(desc.tetBoundMesh.fileName.c_str());
    boundMesh.bindToTets(tetmesh);

    if (desc.tetBoundMesh.material.textures.size()) {
      std::string texName = desc.tetBoundMesh.material.textures[0];
      boundMesh.material = Material::basicWithTexture(texName.c_str());
    }
  }
    
  springs = desc.springs;
  drawTets = desc.drawTets;
}

Scene::Scene(const char * tetName,
	     const char * tetBound) 
{
  init();

  tetmesh = TetMesh::ptr(new TetMesh());
  tetmesh->loadobj(tetName);

  if (tetBound) {
    boundMesh.loadobj(tetBound);
    boundMesh.bindToTets(tetmesh);
  }
}

void Scene::init() {
  fps = 0.0;

  simulating = false;
  wireframe = false;
  handleTet = -1;
  handleBary = Vec4f(0, 0, 0, 0);

  drawTets = 1;

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
    
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    
  glEnable(GL_DEPTH_TEST);

  printf("Creating fbo.\n");
  fbo = new FrameBufferObject(1024, 1024);
  printf("Created fbo.\n");
}



void Scene::drawHud() {
  glPushAttrib(GL_ALL_ATTRIB_BITS);

  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);
  glColor3f(1, 1, 1);

  glMatrixMode(GL_PROJECTION); 
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0, view->aspect(), 0, 1, -1, 1);

  glMatrixMode(GL_MODELVIEW); 
  glPushMatrix();
  glLoadIdentity();

  UI::drawString(0.01, 0.01, "fps: %f", fps);

  glMatrixMode(GL_PROJECTION); 
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW); 
  glPopMatrix();
 
  glPopAttrib();
}

/**
   Super quick hack for drawing a projected ground shadow.
*/
void Scene::drawGround() {
  int sep = 25;
  int xstart = -25;
  int zstart = -25;
  int zend = 25;
  int xend = 25;

    
  fbo->bind(1);
  glViewport(0, 0, 1024, 1024);
  glClearColor(1, 1, 1, 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glClearColor(0, 0, 0, 0);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(xstart, xend, zstart, zend, 0, 2*fabs(kGroundHeight));

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  glRotatef(-90, 1, 0, 0);
  glTranslatef(0, -kGroundHeight, 0);

  glColor3f(0, 0, 0);
  boundMesh.draw();
    
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
    
  //nacb::Image8 rend(fbo->getWidth(), fbo->getHeight(), 4);
  //glReadPixels(0, 0, rend.w, rend.h, rend.channelToGL(), rend.typeToGL(), rend.data);
  //rend.write("/tmp/rend.tga");

  fbo->bind(0);
  glViewport(0, 0, view->width(), view->height());

  float xrange = xend - xstart;
  float zrange = zend - zstart;

  glColor3f(1, 1, 1);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, fbo->getColorTexture());

  for (int z=zstart; z<zend; z+=sep) {
    glBegin(GL_QUAD_STRIP);
    glNormal3f(0, 1, 0);
    glTexCoord2f(0, float(z - zstart)/zrange);
    glVertex3f(xstart, kGroundHeight, z);
    glTexCoord2f(0, float(z + sep - zstart)/zrange);
    glVertex3f(xstart, kGroundHeight, z+sep);

    for (int x=xstart + sep; x<=xend; x+=sep) {
      glTexCoord2f(float(x - xstart)/xrange, float(z - zstart)/zrange);
      glVertex3f(x, kGroundHeight, z);
      glTexCoord2f(float(x - xstart)/xrange, float(z + sep - zstart)/zrange);
      glVertex3f(x, kGroundHeight, z+sep);
    }
      
    glEnd();
  }
  glDisable(GL_TEXTURE_2D);
}


void Scene::simulate(double dt) {
  static int cnt = 0;
  static double etime = 0;    
  nacb::StopWatch timer;      
        

  std::vector<Tetrahedron> & tets = tetmesh->tets;
  Matrix fext = Matrix::zeros(3*tetmesh->restVert.size(), 1);
  Matrix M = tetmesh->getMassMatrix();
  vector<Vec3f> & restVert = tetmesh->restVert;
  vector<Vec3f> & vert = tetmesh->vert;
        
  // Gravity 
  for (int i=0; i<(int)vert.size(); i++)
    fext[3*i + 1] = -9.81*M[i];
    

  // Springs.
  for (int i=0; i<(int)springs.size(); i++) {
    Spring & spring = springs[i];
    double kSpringWeight = 0.25;

    for (int k=0; k<4; k++) {
      int vi = tets[spring.tet].vi[k];
      Vec3f diff = spring.pos - vert[vi];
	
      for (int h=0; h<3; h++) {
	fext[3*vi+h] += kSpringWeight*spring.bary[k]*diff.data[h]/(dt*dt);
      }
    }
  }

  // This is really a spring.
  if (handleTet >= 0) {
    Vec3f constraintPos = handlePos;// tetmesh->getWorldCoordinates(handleTet, handleBary);

    for (int k=0; k<4; k++) {
      int i = tets[handleTet].vi[k];
      float wt = handleBary.data[k];

      fext[3*i  ] += 0.5*wt*(constraintPos.x - vert[i].x)/(dt*dt);
      fext[3*i+1] += 0.5*wt*(constraintPos.y - vert[i].y)/(dt*dt);
      fext[3*i+2] += 0.5*wt*(constraintPos.z - vert[i].z)/(dt*dt);
    }
  }

  // Constraints take precedence, and overwrite existing forces.
  for (int i=0; i<(int)vert.size(); i++) {
    if (restVert[i].z <= -3.0) {
      fext[3*i  ] = (restVert[i].x - vert[i].x)/(dt*dt);
      fext[3*i+1] = (restVert[i].y - vert[i].y)/(dt*dt);
      fext[3*i+2] = (restVert[i].z - vert[i].z)/(dt*dt);
    }

    if(vert[i].y < kGroundHeight){
      fext[3*i+1] = (kGroundHeight - vert[i].y)/(dt*dt);
    }
  }

  tetmesh->simulate(0.01, fext);

    
  boundMesh.deform();

  const std::vector<TetMeshFractureInfo> & recent = tetmesh->getRecentFractures();
    
  if (recent.size()) {
    for (int i=0; i<recent.size(); i++)
      recordedFractures.push_back(recent[i]);
      
    std::ofstream ofs("/tmp/fractures.txt");
    boost::archive::text_oarchive oa(ofs);
    oa << (const std::vector<TetMeshFractureInfo> &) recordedFractures;
  }

  boundMesh.fracture();
  boundMesh.splitSuperfluousEdges();    


  // Keep track of simulation time.
  etime += double(timer);
  cnt ++;
            
#ifdef DEBUG_FRACTURE
  //      simulating = !frac;
#endif

  if (cnt % 30 == 0) {
    fps = 30.0/etime;
    etime = 0;
    cnt = 0;
  }
}


void Scene::draw(){
  if (wireframe)
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  else
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  if (drawTets)
    tetmesh->draw();

  glColor3f(0.9, 0.9, 0.9);    
  boundMesh.draw();
  glDisable(GL_TEXTURE_2D);

  // Rest of scene.
  drawGround();

  for (int i=0; i<springs.size(); i++) {
    springs[i].draw(tetmesh->getWorldCoordinates(springs[i].tet, springs[i].bary));
  }
  if (handleTet >= 0)
    Spring(handleTet, handleBary, handlePos).draw(tetmesh->getWorldCoordinates(handleTet,
									       handleBary));

  drawHud();    
}


void Scene::reset() {
  tetmesh->resetToRest();
  boundMesh.deform();
}


void Scene::convertHandleToSpring() {
  if (handleTet >= 0)
    springs.push_back(Spring(handleTet, handleBary, handlePos));
}


bool Scene::handleKeyboard(unsigned char c, int x, int y){
  switch(c){
  case 'p':
      
    {
      static int loaded = 0;
      static int index = 0;
      static std::vector<TetMeshFractureInfo>  frac;
      if (!loaded) {
	try {
	  std::ifstream ifs("/tmp/fractures_read.txt");
	  boost::archive::text_iarchive ia(ifs);
	  ia >> frac;

	  printf("read %d fractures.\n", frac.size());
	}
	catch (std::exception & e) {
	  std::cout << e.what();
	}
	loaded = true;
      }
      if (index < frac.size()) {
	tetmesh->fractures.clear();
	tetmesh->debugFracture(frac[index++]);
	boundMesh.deform();
	boundMesh.fracture();
      }	
    }
    break;
      
  case 'n':
    boundMesh.drawNumbers = !boundMesh.drawNumbers;
    break;

  case 'R':
    reset();
    break;

  case 't':
    tetmesh->fracThresh = 1000000000;
    break;

  case 'r':
    tetmesh->fracThresh = 1000;
    break;

  case 'b':
    boundMesh.drawBinding = !boundMesh.drawBinding;
    break;

  case 'k':
    {
      Quaternion quat = Quaternion::rod(Vec3d(0.1, 0.0, 0));
      view->cquat = view->cquat * quat;
    };
    break;

  case 'l':
    {
      Quaternion quat = Quaternion::rod(Vec3d(0, 0.1, 0));
      view->cpos = quat.rotate(view->cpos);
      view->cquat = quat * view->cquat;
    };
    break;
    
  case 'c':
    convertHandleToSpring();
    break;

  case '1':
    drawTets = !drawTets;
    break;

  case 'w':
    wireframe = !wireframe;
    break;

  case 's':
    simulating = !simulating;
    break;

  case ' ':
    simulate(0.01);
    break;
  }
}


void Scene::moveHandleForView(InteractiveCameraView * view, int x, int y) {
  double mod[16], proj[16];
  GLint vp[4];

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  view->applyProjection();

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  view->applyModelview();

  glGetIntegerv(GL_VIEWPORT, vp);
  glGetDoublev(GL_MODELVIEW_MATRIX, mod);
  glGetDoublev(GL_PROJECTION_MATRIX, proj);
    
  Vec3f handle = handlePos; //tetmesh->getWorldCoordinates(handleTet, handleBary);
    
  double ox = handle.x, oy = handle.y, oz = handle.z;
  double winx, winy, winz;
  gluProject(ox, oy, oz, mod, proj, vp, &winx, &winy, &winz);
    
    
  winx = x;
  winy = view->height() - y;
    
  gluUnProject(winx, winy, winz, mod, proj, vp, &ox, &oy, &oz);
    
  handlePos = Vec3f(ox, oy, oz);
}


void Scene::selectHandleForView(InteractiveCameraView * view, int button, int down, int x, int y) {
  if(button == 0 && down){
    GLuint buffer[1024];
      
    glSelectBuffer(1024, buffer);
      
    glRenderMode(GL_SELECT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
      
    GLint vp[4] = {0, 0, view->width(), view->height()};
    gluPickMatrix(x, view->height() - y, 3, 3, vp);
    view->applyProjection();
      
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    view->applyModelview();
      
    glInitNames();

    glPushName(0);
    tetmesh->drawSelect();

    int nhits = glRenderMode(GL_RENDER);
    GLuint * b = buffer;
    int neari = -1;
    double neardepth = 10;


    Mat4x4 mod, proj;
	
    glGetIntegerv(GL_VIEWPORT, vp);
    glGetDoublev(GL_MODELVIEW_MATRIX, mod.data);
    glGetDoublev(GL_PROJECTION_MATRIX, proj.data);
      
    double winx = view->width()/2, winy = view->height()/2, winz = 1.0;
    double ox, oy, oz;
    double cx, cy, cz;
      
    gluUnProject(winx, winy, winz, mod.data, proj.data, vp, &ox, &oy, &oz);
    winz = 0.0001;
    gluUnProject(winx, winy, winz, mod.data, proj.data, vp, &cx, &cy, &cz);
    
    Vec3f cc, farPoint;

    cc = Vec3f(cx, cy, cz);
    farPoint = Vec3f(ox, oy, oz);

    for(int i=0; i<nhits; i++){
      int nnames = *b;
      double mn = double(b[1])/((unsigned int)0xFFFFFFFF);
      double mx = double(b[2])/((unsigned int)0xFFFFFFFF);
      int names[2] = {b[3], b[4]};
	
      // printf("hit: %d,%d  %lf %lf\n", names[0], names[1], mn, mx);
      b += 3 + nnames;

      double t = 0;
      if(mn < neardepth && names[0] < tetmesh->tets.size() && 
	 tetmesh->tets[names[0]].intersectSegment(tetmesh->vert, cc, farPoint, t)){
	neardepth = mn;
	neari = names[0];
      }
    }

    handleTet = neari;
      
    if (handleTet >= 0) {
	
      double t = 0;
      if (tetmesh->tets[handleTet].intersectSegment(tetmesh->vert, cc, farPoint, t)) {
	printf("intersects tet: %f\n", t);
      }
	
      Vec3f coords = cc * (1.0 - t) + farPoint * t;
      handleBary = tetmesh->getBaryCoordinates(handleTet, coords);
      handlePos = coords;

      for (int k=0; k<4; k++)
	if (handleBary.data[k] >= 0.8) {
	  handleBary = Vec4f(k==0, k==1, k==2, k==3);
	  break;
	}

      farPoint = coords;
      printf("bary: %f %f %f %f\n", handleBary.x, handleBary.y, handleBary.z, handleBary.w);
    }
  }
  else {
    handleTet = -1;
  }
}

void Scene::removeSpringForView(InteractiveCameraView * view, int button, int down, int x, int y) {
  if(button == 0 && down){
    GLuint buffer[1024];
      
    glSelectBuffer(1024, buffer);
      
    glRenderMode(GL_SELECT);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
      
    GLint vp[4] = {0, 0, view->width(), view->height()};
    gluPickMatrix(x, view->height() - y, 3, 3, vp);
    view->applyProjection();
      
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    view->applyModelview();
      
    glInitNames();

    glPushName(0);
    for (int i=0; i<springs.size(); i++) {
      glLoadName(i);
      springs[i].draw(tetmesh->getWorldCoordinates(springs[i].tet, springs[i].bary));
    }

    int nhits = glRenderMode(GL_RENDER);
    GLuint * b = buffer;
    int neari = -1;
    double neardepth = 10;

    for (int i=0; i<nhits; i++) {
      int nnames = *b;
      double mn = double(b[1])/((unsigned int)0xFFFFFFFF);
      double mx = double(b[2])/((unsigned int)0xFFFFFFFF);
      int names[2] = {b[3], b[4]};
	
      // printf("hit: %d,%d  %lf %lf\n", names[0], names[1], mn, mx);
      b += 3 + nnames;
      if (mn < neardepth) {
	neardepth = mn;
	neari = names[0];
      }
    }

    if (neari >= 0 && neari < springs.size()) {
      springs.erase(springs.begin() + neari);
    }
  }
}




