#include "SplitContext.h"
#include "TetBoundMesh.h"

typedef TetBoundMesh::TetBinding TetBinding;

SplitContext::SplitContext(TetBoundMesh::TetBoundMesh & m, 
			   TetMeshFractureInfo & f) : mesh(m), frac(f) {
  stride = mesh.vert.size()*10 + 100;
}

void SplitContext::setSeparatingTriangle(const TetMeshFractureInfo::SeparatingTriangle & sep,
					 const TriangleIndices & triInds,
					 const Triangle & triangle) {
  this->sep = sep;
  this->triInds = triInds;
  this->triangle = triangle;
      
}


int SplitContext::getVertexIndex(int vi) {
  assert(splitVertMap.count(vi));
  return splitVertMap[vi];
}

SplitVertex & SplitContext::getVertex(int vi) {
  if (splitVertMap.count(vi)) {
    return splitVerts[splitVertMap[vi]];
  }
  mesh.binding[vi] = TetBinding(sep.tetAbove,
				mesh.tets->getRestBaryCoordinates(sep.tetAbove, 
								  mesh.restVert[vi]));
  int belowIndex = mesh.vert.size();
  mesh.binding.push_back(TetBinding(sep.tetBelow,
				    mesh.tets->getRestBaryCoordinates(sep.tetBelow, 
								      mesh.restVert[vi])));
  mesh.vert.push_back(mesh.vert[vi]);
  mesh.restVert.push_back(mesh.restVert[vi]);
      
  int splitIndex = splitVerts.size();
  splitVertMap[vi] = splitIndex;
  splitVertMap[belowIndex] = splitIndex;

  splitVerts.push_back(SplitVertex(vi, belowIndex));
  return splitVerts[splitIndex];
}

int SplitContext::getEdgePointIndex(EdgePoint & edgePoint) {
  int index = edgePoint.v1 * stride + edgePoint.v2;
  assert(edgePointMap.count(index));
  return edgePointMap[index];
}

SplitVertex & SplitContext::getEdgePoint(EdgePoint & edgePoint) {
  int index1 = edgePoint.v1 * stride + edgePoint.v2;
  int index2 = edgePoint.v2 * stride + edgePoint.v1;
  if (edgePointMap.count(index1) ){
    return splitVerts[edgePointMap[index1]];
  }

  int splitIndex = splitVerts.size();
  edgePointMap[index1] = splitIndex;
  edgePointMap[index2] = splitIndex;

  int vi1 = edgePoint.v1;
  int vi2 = edgePoint.v2;
  double t = edgePoint.t;

  int v1 = mesh.vert.size();
  int v2 = mesh.vert.size();
      
  // Add two extra points (after each other; one above plane, one below).
  Vec3f newVert, newRestVert;
  newRestVert = mesh.restVert[vi1]*(1.0 - t) + mesh.restVert[vi2]*t;
  newVert = mesh.vert[vi1]*(1.0 - t) + mesh.vert[vi2]*t;


  mesh.binding.push_back(TetBinding(sep.tetAbove,
				    mesh.tets->getRestBaryCoordinates(sep.tetAbove, 
								      newRestVert)));
  mesh.vert.push_back(newVert);
  mesh.restVert.push_back(newRestVert);

  if (!edgePoint.isFaceEdgePoint  ||
      edgePoint.faceEdgeSplitting){	
    v2 = mesh.vert.size();
	
    mesh.binding.push_back(TetBinding(sep.tetBelow,
				      mesh.tets->getRestBaryCoordinates(sep.tetBelow, 
									newRestVert)));      
	
    mesh.vert.push_back(newVert);     
    mesh.restVert.push_back(newRestVert);
	
    if (edgePoint.isFaceEdgePoint) {
      printf("Created face edge point: %d %d\n", v1, v2);
    }
  }
  else {
    printf("Edge point is face edge on non-splitting edge.\n");
  }

  splitVertMap[v1] = splitIndex;
  splitVertMap[v2] = splitIndex;

  SplitVertex splitVertex(v1, v2);
  splitVerts.push_back(splitVertex);
  return splitVerts[splitIndex];
}
    
bool SplitContext::triangleEdgeSplitting(EdgePoint & facePoint) {
  bool splitting = (facePoint.v1 == frac.vold) || (facePoint.v2 == frac.vold);
  return splitting;
}

// FIXME: should this actually create and initialize a split point? 
SplitVertex SplitContext::createFacePoint(EdgePoint & facePoint) {
  double t = facePoint.t;
  Vec3f coords = 
    mesh.tets->vert[facePoint.v1] * (1.0 - t) + 
    mesh.tets->vert[facePoint.v2] * t;
  Vec3f rest = mesh.tets->restVert[facePoint.v1] * (1.0 - t) + 
    mesh.tets->restVert[facePoint.v2] * t;

  Vec4f bary = mesh.tets->getRestBaryCoordinates(sep.tetAbove, rest);
      
  // sep_edge_points[facePointEdge[0]].push_back(vert.size());
  // vert_to_added[vert.size()] = added_points.size();
  // added_by_this_face.push_back(added_points.size());
  // added_points.push_back(AddedPoint(vert.size(), trii, -1, -1, false));

  int v1 = mesh.vert.size();
  int v2 = mesh.vert.size();
      
  mesh.binding.push_back(TetBinding(sep.tetAbove, bary));
  mesh.restVert.push_back(rest);
  mesh.vert.push_back(coords);
      
      
      
  // If this is a splitting edge, need to add the point 
  // (actually need to cache this).
  if (triangleEdgeSplitting(facePoint) ) {
    v2 = mesh.vert.size();
    bary = mesh.tets->getRestBaryCoordinates(sep.tetBelow, rest);
	
    mesh.binding.push_back(TetBinding(sep.tetBelow, bary));
    mesh.restVert.push_back(rest);
    mesh.vert.push_back(coords);

    printf("Creating face point (split) %d, %d\n", v1, v2);
  }
      
  int splitIndex = splitVerts.size();
  splitVerts.push_back(SplitVertex(v1, v2));
      
  splitVertMap[v1] = splitIndex;
  splitVertMap[v2] = splitIndex;

  return splitVerts[splitIndex];
}

void SplitContext::splitTwoEdgePoints(int trii, 
				      EdgePoint & edgePoint0, 
				      EdgePoint & edgePoint1,
				      int edgeIndex0, int edgeIndex1) {
  SplitVertex  e0 = getEdgePoint(edgePoint0);
  SplitVertex  e1 = getEdgePoint(edgePoint1);

  // e0, e1, and either one or two verts make up the triangle.
  int vi[3] = {mesh.tris[trii].vi[0],
	       mesh.tris[trii].vi[1],
	       mesh.tris[trii].vi[2]};

  // FIXME: normals and texture coordinates...
  // The first vertex in the edge pair is on the top (>= 0)
  int downs[3] = {triangle.above(mesh.restVert[mesh.tris[trii].vi[0]]),
		  triangle.above(mesh.restVert[mesh.tris[trii].vi[1]]),
		  triangle.above(mesh.restVert[mesh.tris[trii].vi[2]])};
  int down = downs[edgeIndex0];
      
  mesh.tris[trii].vi[0] = vi[edgeIndex0];
  mesh.tris[trii].vi[1] = e0.getVertex(down);
  mesh.tris[trii].vi[2] = e1.getVertex(down);

  if (down)
    getEdgePoint(edgePoint0).next = getEdgePointIndex(edgePoint1);
  else
    getEdgePoint(edgePoint1).next = getEdgePointIndex(edgePoint0);

  vector<Mesh::Tri> & tris = mesh.tris;

  if (edgeIndex1 == (edgeIndex0 + 1)) {
    if (downs[edgeIndex0] != downs[(edgeIndex0 + 2) %3]){
      printf("Down conflict 0\n");
    }
    if (downs[edgeIndex0] == downs[(edgeIndex0 + 1) %3]){
      printf("Down conflict 1\n");
    }
	
    // The first triangle is on the same side as above
    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = vi[edgeIndex0];
    tris.back().vi[1] = e1.getVertex(down);
    tris.back().vi[2] = vi[(edgeIndex0 + 2) % 3];
	
    // THis one is on the opposite side.
    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = e0.getVertex(!down);
    tris.back().vi[1] = vi[(edgeIndex0 + 1) % 3];
    tris.back().vi[2] = e1.getVertex(!down);
  }
  else {
    if (downs[(edgeIndex0 + 1) %3 ] == downs[edgeIndex0]){
      printf("Down conflict 2\n");
    }
    if (downs[(edgeIndex0 + 2) %3 ] == downs[edgeIndex0]){
      printf("Down conflict 3\n");
    }
    if (downs[(edgeIndex0 + 2) %3 ] != downs[(edgeIndex0 + 1) % 3]){
      printf("Down conflict 4\n");
    }
	
    // The other triangles are on the other side.
    tris.push_back(Mesh::Tri());		
    tris.back().vi[0] = e0.getVertex(!down);
    tris.back().vi[1] = vi[(edgeIndex0 + 1)%3];
    tris.back().vi[2] = vi[(edgeIndex0 + 2)%3];
	
    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = e0.getVertex(!down);
    tris.back().vi[1] = vi[(edgeIndex0 + 2)%3];
    tris.back().vi[2] = e1.getVertex(!down);
  }

  dontModifySide.insert(trii);
  dontModifySide.insert(tris.size() - 2);
  dontModifySide.insert(tris.size() - 1);
}

void SplitContext::splitOneVertexOneEdge(int trii, int vertIndex,
					 EdgePoint & edgePoint, int edgeIndex) {
  vector<Mesh::Tri> & tris = mesh.tris;
  SplitVertex vert = getVertex(tris[trii].vi[vertIndex]);
  SplitVertex e = getEdgePoint(edgePoint);
  TriangleIndices vi(tris[trii].vi);
            
  int down = triangle.above(mesh.restVert[vi[(edgeIndex + 1) % 3]]);
  printf("OneVertexOneEdge (split) %d, %d (down: %d, using %d)\n", 
	 vert.above, vert.below, down, vi[(edgeIndex + 1) % 3]);
      
  printf("  %d (%d) %d (%d) %d (%d)\n", tris[trii].vi[0], 
	 triangle.above(mesh.restVert[vi[0]]),
	 tris[trii].vi[1], 
	 triangle.above(mesh.restVert[vi[1]]),
	 tris[trii].vi[2],
	 triangle.above(mesh.restVert[vi[2]]));

  if (down)
    getVertex(tris[trii].vi[vertIndex]).next = getEdgePointIndex(edgePoint);
  else
    getEdgePoint(edgePoint).next = getVertexIndex(tris[trii].vi[vertIndex]);

  tris[trii].vi[0] = vert.getVertex(down);
  tris[trii].vi[1] = e.getVertex(down);
  tris[trii].vi[2] = vi[(edgeIndex + 1) % 3];

  tris.push_back(Mesh::Tri());
  tris.back().vi[0] = vert.getVertex(!down);
  tris.back().vi[1] = vi[edgeIndex];
  tris.back().vi[2] = e.getVertex(!down);      

  dontModifySide.insert(trii);
  dontModifySide.insert(tris.size() - 1);
}

void SplitContext::splitOneEdgeOneFace(int trii,
				       EdgePoint & edgePoint, 
				       int edgeIndex,
				       EdgePoint & facePoint,
				       int faceEdgeIndex) {
  vector<Mesh::Tri> & tris = mesh.tris;
  // Keep track of the top and bottom points so that we can
  // use them later (in the next separating triangle that shares the
  // edge that caused this facePoint.
  TriangleIndices vi(tris[trii].vi);
  int downs[3] = {triangle.above(mesh.restVert[vi[edgeIndex]]),
		  triangle.above(mesh.restVert[vi[(edgeIndex + 1) % 3]]),
		  triangle.above(mesh.restVert[vi[(edgeIndex + 2) % 3]])};
  // Here
  int down = downs[0];
  SplitVertex e = getEdgePoint(edgePoint);
  {
    bool hasDown = false;
    SplitVertex c = createFacePoint(facePoint);

    // Update the triangles.
    // v1-e-c, e'-v2-c', (down determined by the pair) v2-v3-c, v3-v1-c
    // in the last two disagreeing pairs will be by default connected to the top.		
    tris[trii].vi[0] = vi[edgeIndex];
    tris[trii].vi[1] = e.getVertex(down);
    tris[trii].vi[2] = c.getVertex(down);

    printf("splitOneEdge, edge: %d %d,  Face: %d, %d\n", e.above, e.below, c.above, c.below);

    if (downs[0] == downs[1]) {
      printf("Inconsistent downs.\n");
    }

    if (down)
      getEdgePoint(edgePoint).next = getVertexIndex(c.above);
    else
      getVertex(c.above).next = getEdgePointIndex(edgePoint);
	 

    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = e.getVertex(!down);
    tris.back().vi[1] = vi[(edgeIndex + 1) % 3];
    tris.back().vi[2] = c.getVertex(!down);
	 
    // Use the connected vertices to make a best guess of which
    // point (top or bottom) these triangles should be connected to.
    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = vi[(edgeIndex + 1) % 3];
    tris.back().vi[1] = vi[(edgeIndex + 2) % 3];
    tris.back().vi[2] = c.getVertex(downs[1] && downs[2]);
	 
	 
    // Swapping the triangles ensures the triangle that needs to 
    // be split will be considered.
    tris.push_back(Mesh::Tri());
    tris.back().vi[0] = vi[(edgeIndex + 2) % 3];
    tris.back().vi[1] = vi[edgeIndex];
    tris.back().vi[2] = c.getVertex(downs[0] && downs[2]);

    dontModifySide.insert(trii);
    dontModifySide.insert(tris.size() - 3);
    dontModifySide.insert(tris.size() - 2);
    dontModifySide.insert(tris.size() - 1);
  }
}
    
void SplitContext::splitSingleEdge(int trii, EdgePoint & edgePoint, int ei) {
  TriangleIndices vi(mesh.tris[trii].vi);
  SplitVertex edge = getEdgePoint(edgePoint);
      
  int down = triangle.above(mesh.restVert[mesh.tris[trii].vi[(ei + 1) % 3]]);
      
  mesh.tris[trii].vi[0] = edge.getVertex(down);
  mesh.tris[trii].vi[1] = vi[(ei + 1) % 3];
  mesh.tris[trii].vi[2] = vi[(ei + 2) % 3];
      
  mesh.tris.push_back(Mesh::Tri());
  mesh.tris.back().vi[0] = vi[ei];
  mesh.tris.back().vi[1] = edge.getVertex(!down);
  mesh.tris.back().vi[2] = vi[(ei + 2) % 3];
      
  dontModifySide.insert(trii);
  dontModifySide.insert(mesh.tris.size() - 1);
}

void SplitContext::splitSingleVert(int trii, const Vec3<int> & vertOnTri) {
  // if (!context.dontModifySide.count(trii))
  if (1) 
  { 
    vector<Mesh::Tri> & tris = mesh.tris;
    vector<Vec3f> & restVert = mesh.restVert;
    vector<TetBinding> & binding = mesh.binding;
	
    // No adjacencies need to be added.
    for (int k=0; k<3; k++) {
      // Only one vertex is on the face.
      if (vertOnTri.data[k]) {
	int vi = tris[trii].vi[k];		
	SplitVertex splitVert = getVertex(vi);
	    
	int v1 = tris[trii].vi[(k + 1) % 3];
	int v2 = tris[trii].vi[(k + 2) % 3];
	    
	// FIXME: consesus?
	int b1 = triangle.above(restVert[v1]);
	int b2 = triangle.above(restVert[v2]);
	int below = b1;

	if (triangle.distance(restVert[tris[trii].vi[(k + 2) % 3]]) <
	    triangle.distance(restVert[tris[trii].vi[(k + 1) % 3]]))
	  below = b2;
	    
	// This is not gauranteed.
	int nabove = frac.tetAboveSet.count(binding[v1].tet) + frac.tetAboveSet.count(binding[v2].tet);
	int nbelow = frac.tetBelowSet.count(binding[v1].tet) + frac.tetBelowSet.count(binding[v2].tet);
	    
	printf("nabove: %d nbelow: %d  (%d, %d, %d)\n", nabove, nbelow, vi, v1, v2);
	    
	// Only change if there is consensus
	//if (b1 == b2) 
	{
	  Vec3f v1 = restVert[tris[trii].vi[(k + 1) % 3]], v2 = restVert[tris[trii].vi[(k + 2) % 3]];
	  double d = (triangle.n.dot(v1) + triangle.d)  + (triangle.n.dot(v2) + triangle.d);
	  below = d>0;
	  printf("below: %d %d  %f %d\n", b1, b2, d, below);
	      
	  //tris[trii].vi[k] = splitVert.getVertex(below);
	}
	tris[trii].vi[k] = splitVert.getVertex(nbelow == 2);
	break;
      }
    }
  }
}

void SplitContext::clearNext() {
  for (int i=0; i<splitVerts.size(); i++) {
    splitVerts[i].next = -1;
  }
}

void SplitContext::findPath() {
  std::set<int> visited;
      
  for (int i=0; i<splitVerts.size(); i++) {
    if (!visited.count(i)) {
      int finger = i;
      vector<int> path;

      while (finger>=0 && !visited.count(finger)) {	    
	visited.insert(finger);

	path.push_back(finger);

	finger = splitVerts[finger].next;
      }
	  
      bool loop = (path.size() && splitVerts[path.back()].next == i);
      printf("Path length: %d is loop: %d\n", path.size(), loop);
      if (loop) {
	for (int down=0; down<=1; down++) {
	  for (int i=0; i<path.size(); i++)
	    printf("%d ", splitVerts[path[i]].getVertex(down));
	  printf("\n");
	}
	    
	for (int i=1; i+1<path.size(); i++) {
	  ignoreTris.insert(mesh.tris.size());

	  mesh.tris.push_back(Mesh::Tri());
	  mesh.tris.back().vi[0] = splitVerts[path[0]].getVertex(0);
	  mesh.tris.back().vi[1] = splitVerts[path[i]].getVertex(0);
	  mesh.tris.back().vi[2] = splitVerts[path[i+1]].getVertex(0);
	}
	    	    
	// Triangulate the hole (bottom)
	for (int i=1; i+1<path.size(); i++) {
	  ignoreTris.insert(mesh.tris.size());

	  mesh.tris.push_back(Mesh::Tri());
	  mesh.tris.back().vi[0] = splitVerts[path[0]].getVertex(1);
	  mesh.tris.back().vi[2] = splitVerts[path[i]].getVertex(1);
	  mesh.tris.back().vi[1] = splitVerts[path[i+1]].getVertex(1);
	}
      }
    }
  }      
}

SplitVertex & SplitContext::getSplitTriangleVertex(int vi) {
  if (splitTriangleVertMap.count(vi)) 
    return splitVerts[splitTriangleVertMap[vi]];

  int splitIndex = splitVerts.size();
  int v1 = mesh.vert.size();
  int v2 = mesh.vert.size();

  mesh.vert.push_back(mesh.tets->vert[vi]);
  mesh.restVert.push_back(mesh.tets->restVert[vi]);
  mesh.binding.push_back(TetBinding(sep.tetAbove,
				    mesh.tets->getRestBaryCoordinates(sep.tetAbove, 
								      mesh.restVert.back())));
      
  if (vi == frac.vold || vi == frac.vnew) {
    v2 = mesh.vert.size();
	
    mesh.vert.push_back(mesh.tets->vert[vi]);
    mesh.restVert.push_back(mesh.tets->restVert[vi]);
    mesh.binding.push_back(TetBinding(sep.tetBelow,
				      mesh.tets->getRestBaryCoordinates(sep.tetBelow, 
									mesh.restVert.back())));
			       
			       
    splitTriangleVertMap[frac.vold] = splitIndex;
    splitTriangleVertMap[frac.vnew] = splitIndex;
  }

  splitTriangleVertMap[vi] = splitIndex;
      
  splitVerts.push_back(SplitVertex(v1, v2));
  return splitVerts[splitIndex];
}

void SplitContext::fixFacePoints() {
  std::vector<int> pointsByEdge[3];

  for (int i=0; i<splitVerts.size(); i++) {
    int edge = triangle.onEdge(mesh.restVert[splitVerts[i].getVertex(false)]);
	
    if (edge >= 0) {
      printf("%d is on edge %d\n", i, edge);
      pointsByEdge[edge].push_back(i);
    }
  }

      
  for (int k=0; k<3; k++) {
    vector<pair<double, int> > points;

    if (mesh.tetVertInside[triInds[k]]) {
      getSplitTriangleVertex(triInds[k]);
      points.push_back(make_pair(0.0, splitTriangleVertMap[triInds[k]]));
    }

    if (mesh.tetVertInside[triInds[(k + 1) % 3]]) {
      getSplitTriangleVertex(triInds[(k + 1) % 3]);
      points.push_back(make_pair(0.0, splitTriangleVertMap[triInds[(k + 1) % 3]]));	  
    }

    Vec3f origin = mesh.tets->restVert[triInds[k]];
    Vec3f dir = mesh.tets->restVert[triInds[(k + 1) % 3]] - origin;
    dir.normalize();

    for (int h=0; h<pointsByEdge[k].size(); h++) {
      int si = pointsByEdge[k][h];
      Vec3f diff = mesh.restVert[splitVerts[si].getVertex(0)] - origin;
      double t = diff.dot(dir);
	  
      points.push_back(make_pair(t, si));
    }

    std::sort(points.begin(), points.end());
    for (int k=0; k+1<points.size(); k+=2) {
      printf("linnking %d - %d\n", points[k].second, points[k+1].second);
      splitVerts[points[k].second].next = points[k+1].second;
    }
  }
}
