#ifndef SPLIT_CONTEXT_H
#define SPLIT_CONTEXT_H

//#include "TetBoundMesh.h"

#include "TetMesh.h"
#include "TetMeshFractureInfo.h"
#include "Triangle.h"

#include <nmath/vec3.h>
#include <map>
#include <vector>

using namespace std;
using namespace nacb;

class TetBoundMesh;


/**
   An edge point is a point that was created on a bound mesh
   triangle edge.  It maintains the location (linear combination
   parameter) and the vertex indices.
*/
class EdgePoint {
 public:
  bool valid;
  int v1;
  int v2;
  double t;

  bool isFaceEdgePoint;
  int faceEdgeSplitting;

  EdgePoint() {
    v1 = v2 = -1;
    t = 0;
    valid = 0;
    isFaceEdgePoint = false;      
  }
    
  EdgePoint(int _v1, int _v2, double _t) {
    v1 = _v1;
    v2 = _v2;
    t = _t;
    valid = true;
    isFaceEdgePoint = false;
  }

  operator bool () const {
    return valid;
  }

  Vec3f operator()(const std::vector<Vec3f> & vert) const {
    return vert[v1] * (1.0 - t) + vert[v2]*t;
  }

  /**
     Create an edge point from the triangle and the given vertices.
     Edge point is only valid if the intersection between the segment
     and the triangle is valid.
   */
  static EdgePoint createFromTriangle(const Triangle & triangle,
				      const std::vector<Vec3f> & vert,
				      int v1, int v2) {
    double t;
    if (triangle.intersectSegment(vert[v1], vert[v2], t)) {
      return EdgePoint(v1, v2, t);
    }
    return EdgePoint();
  }
};

/**
   A vertex can be split into two vertices (one on top and one below the
   split). The indices and adjacency information are maintained hree.
*/
class SplitVertex {
 public:
 SplitVertex(int _a = -1, int _b = -1) : above(_a), below(_b), next(-1) { }
  
  int getVertex(bool b) const {
    if (b) return below;
    return above;
  }
  
 public:
  int above, below;
  int next;
};


/**
   This class does the book-keeping for fracturing a bound mesh.
   Keeps track of all the vertices that have been split, and also
   the indices of these vertices in the mesh.
*/
class SplitContext {
  TetBoundMesh & mesh;

  std::vector<SplitVertex> splitVerts;

  std::map<int, int> splitTriangleVertMap;

  // This avoids pointers (index into the array).
  std::map<int, int> splitVertMap;
  std::map<int, int> edgePointMap;

  TetMeshFractureInfo::SeparatingTriangle sep;
  TriangleIndices triInds;
  Triangle triangle;
    
  int stride;
    
 public:
  std::set<int> ignoreTris;
  std::set<int> dontModifySide;

 public:
  TetMeshFractureInfo & frac;

  SplitContext(TetBoundMesh & m, TetMeshFractureInfo & f);

  void setSeparatingTriangle(const TetMeshFractureInfo::SeparatingTriangle & sep,
			     const TriangleIndices & triInds,
			     const Triangle & triangle);
  int getVertexIndex(int vi);
  SplitVertex & getVertex(int vi);

  int getEdgePointIndex(EdgePoint & edgePoint);

  SplitVertex & getEdgePoint(EdgePoint & edgePoint);
  bool triangleEdgeSplitting(EdgePoint & facePoint);

  SplitVertex createFacePoint(EdgePoint & facePoint);

  void splitTwoEdgePoints(int trii, 
			  EdgePoint & edgePoint0, 
			  EdgePoint & edgePoint1,
			  int edgeIndex0, int edgeIndex1);

  void splitOneVertexOneEdge(int trii, int vertIndex,
			     EdgePoint & edgePoint, int edgeIndex);
  void splitOneEdgeOneFace(int trii,
			   EdgePoint & edgePoint, 
			   int edgeIndex,
			   EdgePoint & facePoint,
			   int faceEdgeIndex);
  void splitSingleEdge(int trii, EdgePoint & edgePoint, int ei);

  void splitSingleVert(int trii, const Vec3<int> & vertOnTri);

  void clearNext();
  void findPath();

  SplitVertex & getSplitTriangleVertex(int vi);

  void fixFacePoints();
};


#endif // SPLIT_CONTEXT_H
