#include "TetBoundMesh.h"
#include "SplitContext.h"

bool TetBoundMesh::fracture() {
  /**
     Keep track of the vertices that are split (across iterations).
  */
  for (int fraci=0; fraci<(int)tets->fractures.size(); fraci++) {
    TetMeshFractureInfo & frac = tets->fractures[fraci];     
    SplitContext context(*this, frac);
    /*
      For each triangle in the separatingTriangles list,
      we need to find all triangles with edges that intersect it.
      Also need to find all triangles that intersect an edge separating triangle.
      (There should be pairs of these).	
	
      For each new vertex, need to find a path that links them.
      These paths will be linked through the tetrahedra vertices.
      Need to start at some path edge (or the end), and go through
      the separating triangle edges until other path starts are found,
      or until sep-triangle vertics are found.  The closed path should
      be triangulated.	 
    */      
    for (int sepi=0; sepi<(int)frac.sepTris.size(); sepi++) {
      TetMeshFractureInfo::SeparatingTriangle sep = frac.sepTris[sepi];
      TriangleIndices triInds = 
	Tetrahedron::tris[sep.tetAboveTriIndex].mapIndices(tets->tets[sep.tetAbove].vi);
      Triangle triangle(tets->restVert, triInds);

      context.frac = frac;
      context.setSeparatingTriangle(sep, triInds, triangle);

      if (triangle.below(tets->getWorldCoordinates(sep.tetAbove, Vec4f(0.25, 0.25, 0.25, 0.25)))) {
	printf("Tet above is on the wrong side!\n");
      }
      if (triangle.above(tets->getWorldCoordinates(sep.tetBelow, Vec4f(0.25, 0.25, 0.25, 0.25)))) {
	printf("Tet below is on the wrong side!\n");
      }
	
      int ntris = tris.size();
      const double kVertOnTri = 2e-4;
      const double kEdgeOnEdge = 1e-4;
      const double kFaceOnEdge = 1e-4;
      int caseHistogram[10];
      int partialBoundHist[4];

      memset(caseHistogram, 0, sizeof(caseHistogram));
      memset(partialBoundHist, 0, sizeof(partialBoundHist));

      for (int trii=0; trii<ntris; trii++) {	 
	int nbound = 0;
	int nabove = 0;

	if (context.ignoreTris.count(trii)) 
	  continue;

	for (int k=0; k<3; k++) {
	  int v = tris[trii].vi[k];
	  nbound += frac.affectsTetrahedron(binding[v].tet);
	  nabove += frac.tetAboveSet.count(binding[v].tet);
	  // (binding[v].tet == sep.tetAbove) || (binding[v].tet == sep.tetBelow);
	}
	if (!nbound)
	  continue;
	  
	// Not clear if we should try these guys too.
	if (nbound != 3) {
	  partialBoundHist[nbound]++;
	}

	if (nabove == 3 || nabove == 0) { 
	  //printf("No sense splitting (all above or below)\n");
	  //continue;
	}

	Vec3<int> vertOnTri(triangle.distance(restVert[tris[trii].vi[0]]) < kVertOnTri,
			    triangle.distance(restVert[tris[trii].vi[1]]) < kVertOnTri,
			    triangle.distance(restVert[tris[trii].vi[2]]) < kVertOnTri);
	int numVertPoints = vertOnTri.x + vertOnTri.y + vertOnTri.z;

	// Get the edge points.
	EdgePoint edgePoints[3];
	vector<int> edgePointIndices;

	for (int k=0; k<3; k++) {
	  int v1 = tris[trii].vi[k];
	  int v2 = tris[trii].vi[(k + 1)%3];

	  // Edge points can only exist when the neighboring verts are not on septri.
	  if (!vertOnTri.data[k] && !vertOnTri.data[(k + 1) % 3]) {
	    // Special case, when edges intersect (like a face point).
	    double t = 0;

	    triangle.intersectSegment(restVert[v1], restVert[v2], t);

	    if (t > 0 && t < 1) {
	      Vec3f co = restVert[v1]*(1.0 - t) + restVert[v2]*t;
	      int edge = triangle.onEdge(co, kEdgeOnEdge);

	      if (edge >= 0) {
		EdgePoint facePoint(triInds[edge], triInds[(edge + 1) % 3], 0);
		edgePoints[k] = EdgePoint(v1, v2, t);
		edgePoints[k].isFaceEdgePoint = true;
		edgePoints[k].faceEdgeSplitting = context.triangleEdgeSplitting(facePoint);
	      }
	    }
	      
	    if (!edgePoints[k])
	      edgePoints[k] = EdgePoint::createFromTriangle(triangle, restVert, v1, v2);
	      
	    if (edgePoints[k])
	      edgePointIndices.push_back(k);
	  }
	}
	int numEdgePoints = edgePointIndices.size();
	  
	// Get the face points.
	EdgePoint facePoints[3];
	vector<int> facePointEdge;
	int numFacePoints = 0;

	Triangle triangle2(restVert, tris[trii].vi);
	for (int k=0; k<3; k++) {
	  int v1 = triInds[k];
	  int v2 = triInds[(k + 1) % 3];
	    
	  facePoints[k] = EdgePoint::createFromTriangle(triangle2, tets->restVert, v1, v2);
	  if (facePoints[k]) {
	    Vec3f co = facePoints[k](tets->restVert);
	    bool bad = false;

	    for (int j=0; j<3; j++) {
	      if (vertOnTri.data[j] && (co - restVert[tris[trii].vi[j]]).len() < kVertOnTri) {
		bad = true;
		break;
	      }
	      else {
		if ((co - restVert[tris[trii].vi[j]]).len() < kVertOnTri) {
		  printf("Face vert close to point.\n");
		}
	      }
	    }
	    if (!bad) {
	      for (int j=0; j<(int)edgePointIndices.size(); j++) {
		Vec3f coEdge = edgePoints[edgePointIndices[j]](restVert);
		if ((co - coEdge).len() < kFaceOnEdge) {
		  printf("Face vert close to edge point.\n");
		  bad = true;
		  break;
		}
	      }
	    }
	    if (!bad) {
	      if (facePoints[k]) {
		facePointEdge.push_back(k);
		numFacePoints += facePoints[k];
	      }
	    }
	    else {
	      printf("Face point is bad.\n");
	      facePoints[k] = EdgePoint();
	    }
	  }
	}

	if (numVertPoints == 0 && numEdgePoints == 0 && numFacePoints == 0)
	  continue;
	  
	  
	int caseIndex = 0;

	// Case 0
	if (numVertPoints == 1 && !numEdgePoints && !numFacePoints) {
	  context.splitSingleVert(trii, vertOnTri);	    
	  
	  caseIndex = 0;
	}
	// Case 1
	else if (numEdgePoints == 2 && !numVertPoints && !numFacePoints) {
	  caseIndex = 1;

	    	    
	  context.splitTwoEdgePoints(trii, 
				     edgePoints[edgePointIndices[0]], 
				     edgePoints[edgePointIndices[1]],
				     edgePointIndices[0], edgePointIndices[1]);	    
	}
	// Case 2
	else if (!numVertPoints && numEdgePoints == 1 && numFacePoints == 1) {
	  caseIndex = 2;

	  context.splitOneEdgeOneFace(trii, edgePoints[edgePointIndices[0]], edgePointIndices[0],
				      facePoints[facePointEdge[0]], facePointEdge[0]);
	}
	// Case 3
	else if (numVertPoints == 1 && numEdgePoints == 1 && numFacePoints == 0) {
	  int vertIndex = 0;
	  for (int k=0; k<3; k++) {
	    if (vertOnTri.data[k]) {
	      vertIndex = k;
	      break;
	    }
	  }
	  context.splitOneVertexOneEdge(trii, vertIndex,
					edgePoints[edgePointIndices[0]], edgePointIndices[0]);
	  caseIndex = 3;
	}
	// Case 4
	else if (numVertPoints == 1 && numEdgePoints == 0 && numFacePoints == 1) {	    
	  printf(" Case 4 not handled\n");
	  caseIndex = 4;
	}
	// Case 5
	else if (numVertPoints == 2 && !numEdgePoints && !numFacePoints) {
	  int verts[2] = {0, 0};
	  int verti = 0;
	  int below = 0;

	  for (int k=0; k<3; k++) {
	    if (vertOnTri.data[k]) {
	      verts[verti] = k;
	      verti++;
	    }
	    else {
	      below = triangle.above(restVert[tris[trii].vi[k]]);
	    }
	  }
	  tris[trii].vi[verts[0]] = context.getVertex(tris[trii].vi[verts[0]]).getVertex(below);
	  tris[trii].vi[verts[1]] = context.getVertex(tris[trii].vi[verts[1]]).getVertex(below);

	  context.dontModifySide.insert(trii);
	  caseIndex = 5;
	}
	// Case 6
	else if (!numVertPoints && !numEdgePoints && numFacePoints == 2) {
	  caseIndex = 6;
	}
	// Case 7 (single edge)
	else if (!numVertPoints && numEdgePoints == 1 && !numFacePoints) {
	  context.splitSingleEdge(trii, edgePoints[edgePointIndices[0]], edgePointIndices[0]);

	  caseIndex = 7;
	}
	else {
	  isect.push_back(vert[tris[trii].vi[0]]);
	  isect.push_back(vert[tris[trii].vi[1]]);
	  isect.push_back(vert[tris[trii].vi[2]]);

	  caseIndex = 8;
	  printf("Unknown: (vert, edge, face): %d %d %d\n", 
		 numVertPoints, numEdgePoints, numFacePoints);
	}
	caseHistogram[caseIndex]++;
      }

      // Find the new points that are on the edges, sort and fix up adjacency.
      context.fixFacePoints();

      printf("hist: = ");
      for (int k=0; k<9; k++) {
	printf(" %d", caseHistogram[k]);
      }
      printf("\n");

      printf("partial: = ");
      for (int k=0; k<4; k++)
	printf(" %d", partialBoundHist[k]);
      printf("\n");
	
      context.findPath();
      context.clearNext();
    }

    // Handle any of the edges.
    for (int sepi=0; sepi<(int)frac.sepEdges.size(); sepi++) {
      TetMeshFractureInfo::SeparatingEdge sep = frac.sepEdges[sepi];
      for (int vi=0; vi<(int)restVert.size(); vi++) {
	if (sep.onEdge(tets->restVert, restVert[vi], 1e-5) ){
	  printf("%d on sep edge.\n", vi);
	}
      }
    }
  }

 #ifdef MOVE_COLLISION_VERT_INSIDE
  vector<Vec3f> modVert = tets->restVert;
  
  for (int fraci=0; fraci<tets->fractures.size(); fraci++) {
    TetMeshFractureInfo & frac = tets->fractures[fraci];     
    
    for (int sepi=0; sepi<frac.sepTris.size(); sepi++) {
      TetMeshFractureInfo::SeparatingTriangle sep = frac.sepTris[sepi];
      TriangleIndices triInds = 
	Tetrahedron::tris[sep.tetAboveTriIndex].mapIndices(tets->tets[sep.tetAbove].vi);
      Triangle triangle(tets->restVert, triInds);
      
      modVert[frac.vold] -= triangle.n*1e-2;
      modVert[frac.vnew] += triangle.n*1e-2;
    }
  }
  
#ifdef MOVE_TET_VERT
  tets->restVert = modVert;
#else
  // Get the coordinates with the displaced vert, and then rebind to the
  // original tet coords.
  for (int i=0; i<vert.size(); i++) {
    restVert[i] = tets->tets[binding[i].tet].getWorldCoordinates(modVert,
								 binding[i].bary);
    binding[i].bary = tets->getRestBaryCoordinates(binding[i].tet, restVert[i]);
  }
#endif
#endif

    
  initNormalsAutoSmooth(60.0);
  return true;
}
