#ifndef TET_BOUND_MESH_H
#define TET_BOUND_MESH_H

#include "TetMesh.h"
#include "UI.h"

#include <ik/mesh.h>

#include <map>
#include <vector>
#include "Material.h"

using namespace nacb;
using namespace std;

class SplitContext;



/**
   This is a mesh that is attached to a tetrahedral mesh.  Has the
   ability to fracture (given the TetMesh fractures) and it deforms
   with the TetMesh.
*/
class TetBoundMesh : public Mesh {
public:
  class TetBinding {
  public:
    int    tet;
    Vec4f bary;

    TetBinding(int _tet = 0, const Vec4f & _bary = Vec4f()) :
      tet(_tet), bary(_bary) {
      
    }
  };

public:

  TetBoundMesh() {
    drawBinding = false;
    material = Material::ptr(new Material());
  }  

  void bindToTets(const TetMesh::ptr & tets) {
    // Find the closest tetra-hedron in the tetmesh for 
    // every vertex in this.  Assign the vertex to the
    // bary-centric coordinates of the corresponding tet.
    this->tets = tets;
    binding.clear();

    for (int i=0; i<(int)vert.size(); i++) {
      int closest = tets->getClosestTetrahedron(vert[i]);
      Vec4f co = tets->getBaryCoordinates(closest, vert[i]);
      Vec3f recon = tets->getWorldCoordinates(closest, co);

      double n = std::min(std::min(co.x, co.y), std::min(co.y, co.z));
      if (n < 0) {
	printf("%f %f %f %f  dist:%f\n", co.x, co.y, co.z, co.w, (recon - vert[i]).len());
	tets->getClosestTetrahedron(vert[i], true);
      }
      binding.push_back(TetBinding(closest, co));
    }

    // FIXME: this could be sped up. Brute force checking for whether
    //        a tet-vert is inside the mesh.
    tetVertInside = vector<bool>(tets->vert.size());
    vector<Triangle> triangles(tris.size());
    for (int i=0; i<(int)tris.size(); i++) {
      triangles[i] = Triangle(restVert, tris[i].vi);
    }
    
    for (int i=0; i<(int)tets->restVert.size(); i++) {
      Vec3f center = tets->restVert[i];
      Vec3f dir(1, 0, 0);
      bool inside = false;

      double mint = 1e10;
      int bestri = -1;
      for (int k=0; k<(int)triangles.size(); k++) {
	double t;
	if (triangles[k].intersectRay(center, dir, t, 1e-5)) {
	  if (t >= 0 && t <= mint) {
	    bestri = k;
	    mint = t;
	  }
	}
      }
      
      if (bestri >= 0) {
	inside = (dir.dot(triangles[bestri].n) >= 0);
	printf("vert[%d]: ray intersects, inside: %d\n", i, inside);
      }
      tetVertInside[i] = inside;
    }
  }

  void splitSuperfluousEdges() {
    vector<bool> crossTetBound(tets->tets.size(), false);
    for (int i=0; i<(int)tris.size(); i++) {
      if (binding[tris[i].vi[0]].tet != binding[tris[i].vi[1]].tet ||
	  binding[tris[i].vi[0]].tet != binding[tris[i].vi[2]].tet ||
	  binding[tris[i].vi[1]].tet != binding[tris[i].vi[2]].tet) {
	for (int k=0; k<3; k++)
	  crossTetBound[binding[tris[i].vi[k]].tet] = true;
      }
    }

    vector<int> vcount(tets->vert.size(), 0);
    for (int i=0; i<(int)tets->tets.size(); i++) {
      for (int k=0; k<4; k++)
	vcount[tets->tets[i].vi[k]]++;
    }

    for (int i=0; i<(int)tets->tets.size(); i++) {
      if (!crossTetBound[i]) {
	bool split = false;

	for (int k=0; k<4; k++) {
	  if (vcount[tets->tets[i].vi[k]] > 1) {
	    int newIndex = tets->vert.size();
	    int oldIndex = tets->tets[i].vi[k];
	    tets->vert.push_back(tets->vert[oldIndex]);
	    tets->restVert.push_back(tets->restVert[oldIndex]);
	    tets->vs.push_back(tets->vs[oldIndex]);
	    tets->tets[i].vi[k] = newIndex;
	    vcount[oldIndex]--;
	    vcount.push_back(1);
	    split = true;
	  }
	}

	if (split)
	  printf("Can separate a tetrahedron: %d\n", i);
      }
    }	   
  }

  bool fracture();

  bool fracture2();

  void deform() {
    for (int i=0; i<(int)vert.size(); i++) {
      vert[i] = tets->getWorldCoordinates(binding[i].tet, binding[i].bary);
    }
    initNormals();
  }

  void draw() {
    material->bind();
    Mesh::draw();
    material->unbind();

    if (drawBinding) {
      glColor3f(1, 1, 1);
      glBegin(GL_LINES);
      for (int i=0; i<(int)vert.size(); i++) {
	glVertex3fv(vert[i].data);
	glVertex3fv(tets->getWorldCoordinates(binding[i].tet, Vec4f(.25, .25, .25, .25)).data);
      }
      glEnd();
    }
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glDisable(GL_LIGHTING);

    glColor3f(1, 0, 0);
    glPointSize(4);
    glBegin(GL_POINTS);
    for (int i=0; i<(int)isect.size(); i++) {
      glVertex3fv(isect[i].data);
    }    
    glEnd();

    if (drawNumbers) {
      glColor3f(1, 1, 1);
      for (int i=0; i<(int)vert.size(); i++) {
	glPushMatrix();
	glTranslatef(vert[i].x, vert[i].y, vert[i].z);
	UI::drawString(0, 0, "%d", i);
	glPopMatrix();
      }
    }

    glPopAttrib();
  }
  
  bool drawBinding;
  bool drawNumbers;
  Material::ptr material;
    
protected:
  TetMesh::ptr tets;
  
  vector<TetBinding> binding;  
  vector<bool> tetVertInside;

  vector<Vec3f> isect; // !<Debug points to draw.

  friend class SplitContext;
};


#endif // TET_BOUND_MESH_H
