#include "TetMesh.h"
#include <nmisc/timer.h>
#include <nmath/sparsematrix.h>


TetMesh::TetMesh (const Tetrahedron::Material & mat) {
  defaultMaterial = mat;
  fracThresh = 1000;
}

int TetMesh::loadobj(const char * fname){
  printf("loading obj: %s\n", fname);
  int ret = Mesh::loadobj(fname);

    
  buildTetrahedron();
  updateTetVolume();

  tris.clear();
  return ret;
}

  
void TetMesh::removeDuplicateVert(){
  // First remove any duplicates
  const double eps = 1e-5;
  std::vector<int> vmap(vert.size());
  std::vector<Vec3f> vnew;
  std::vector<Vec3f> vrest_new;

  // Could be sped up with a hashing.
  for (int i=0; i<(int)vert.size(); i++) {
    int found = -1;

    for(int j=0; j<(int)vnew.size(); j++){
      Vec3f diff = vnew[j] - vert[i];
      double dist = diff.len();
	
      if(dist < eps){
	found = j;
	break;
      }
    }

    if(found == -1){
      vmap[i] = found = vnew.size();
	
      vnew.push_back(vert[i]);
      vrest_new.push_back(restVert[i]);
    }

    vmap[i] = found;
  }

  for (int i=0; i<(int)tris.size(); i++) {
    for(int k=0; k<3; k++)
      tris[i].vi[k] = vmap[tris[i].vi[k]];
  }

  for(int i=0; i<(int)tets.size(); i++){
    for(int k=0; k<4; k++)
      tets[i].vi[k] = vmap[tets[i].vi[k]];
  }


  printf("removed %d/%d\n", (int)vnew.size(), (int)vert.size());
    
  vert = vnew;
  restVert = vrest_new;
}

// Assume that file format could not represent the 
// tetrahedron (so merge any vertices that are connected).
void TetMesh::buildTetrahedron(){
  tets.clear();
    
  //removeDuplicateVert();

  std::set<int> faceExists;
  std::vector<std::set<int> > vertAdj(vert.size());

  for(int i=0; i<(int)tris.size(); i++){
    const int * vi = tris[i].vi;

    faceExists.insert(faceHash(vi[0], vi[1], vi[2]));
    faceExists.insert(faceHash(vi[1], vi[2], vi[0]));
    faceExists.insert(faceHash(vi[2], vi[0], vi[1]));

    for(int k=0; k<3; k++){
      int v0 = vi[k];
      int v1 = vi[(k + 1)%3];

      vertAdj[v0].insert(v1);
      vertAdj[v1].insert(v0);
    }
  }

  // A tetrahedron exists between 4 vert if the faces
  // 0-1-2  1-0-3  2-1-3  0-2-3  exist
  std::set<int> tetExists;
    
  for(int i=0; i<(int)tris.size(); i++){
    const int * vi = tris[i].vi;
      
    std::set<int>::iterator it(vertAdj[vi[0]].begin());
    std::set<int>::iterator last(vertAdj[vi[0]].end());
    std::set<int> in_tri(vi, vi + 3);

    // Try 4th vertices that are adjacent to this.
    for(; it != last; ++it){
      int v3 = *it;
	
      if(!in_tri.count(v3)){
	if(faceExists.count(faceHash(vi[1], vi[0], v3)) &&
	   faceExists.count(faceHash(vi[2], vi[1], v3)) &&
	   faceExists.count(faceHash(vi[0], vi[2], v3))){
	    
	  if(!anyTetExists(tetExists, vi[0], vi[1], vi[2], v3)){
	    tetExists.insert(tetHash(vi[0], vi[1], vi[2], v3));

	    tets.push_back(Tetrahedron(vi[0], vi[1], vi[2], v3));
	    tets.back().updateNormals(vert);
	  }
	}
      }
    }
  }
  printf("There are %d tets\n", (int)tets.size());
  removeDuplicateVert();
    
  vs = std::vector<Vec3f>(vert.size(), Vec3f(0, 0, 0));
}
  
void TetMesh::draw(){
  for (int i=0; i<(int)tets.size(); i++) {
    tets[i].draw(vert);
  }
}

void TetMesh::drawSelect(){
  for (int i=0; i<(int)tets.size(); i++) {
    glLoadName(i);
      
    tets[i].drawSelect(vert);
  }
}

Matrix TetMesh::getMassMatrix() const {
  Matrix M = Matrix::ones(vert.size()*3, 1)*10;
  return M;
}

Matrix TetMesh::getDampingMatrix() const {
  const double cdamp = 0.5;
  return getMassMatrix() * cdamp;
}

Matrix TetMesh::getCombinedVertVector() const {
  Matrix x(3*vert.size(), 1);

  for (int i=0; i<(int)vert.size(); i++) {
    for (int j=0; j<3; j++)
      x[3*i+j] = vert[i].data[j];
  }
  return x;
}

Matrix TetMesh::getCombinedVelocityVector() const {
  Matrix v(3*vert.size(), 1);

  for(int i=0; i<(int)vert.size(); i++){
    for(int j=0; j<3; j++)
      v[3*i+j] = vs[i].data[j];
  }
  return v;
}

std::vector<std::set<int> > TetMesh::getVertexAdjacency(){
  std::vector<std::set<int> > vadj(vert.size());
    
  for(int ti=0; ti<(int)tets.size(); ti++){
    for(int j=0; j<4; j++){
      for(int k=0; k<4; k++){
	vadj[tets[ti].vi[j]].insert(tets[ti].vi[k]);
	vadj[tets[ti].vi[k]].insert(tets[ti].vi[j]);
      }
    }
  }
  return vadj;
}


void TetMesh::handleFractures(const vector<Matrix> & Rs){
  bool fracture = !false;
  std::set<int> affected;
	
  fractures.clear();
    
  if (fracture) { 
    for (int ti=0; ti<(int)tets.size(); ti++) {
      if (affected.count(ti)) continue;
	
      Matrix x = tets[ti].getVertexVector(vert);
      Matrix x0 = tets[ti].getVertexVector(restVert);
      Matrix stress = tets[ti].E*tets[ti].B*(Rs[ti].transpose()*x - x0);

      Matrix U, V;
      Matrix S(3, 3);
      for (int k=0; k<3; k++) {
	S(k, k) = stress[k];
      }
      S(1, 0) = S(0, 1) = stress[5];
      S(2, 0) = S(0, 2) = stress[4];
      S(2, 1) = S(1, 2) = stress[3];
      
      S.eigSym(U, V);
	
      if(V[2] >= fracThresh) {	  
	Vec3f n = U.getColumn(2);
	// Choose a vert to split properly (use the split front flag).
	int vi = tets[ti].vi[rand() % 4];
	double d = -n.dot(vert[vi]);

	 
	vector<int> neighs = getTetrahedronContaining(vi);

	// Fix the neighboring tetrahedron
	int nabove = 0;
	for (int i=0; i<(int)neighs.size(); i++) {
	  Vec3f center = tets[neighs[i]].getWorldCoordinates(restVert, Vec4f(0.25, 0.25, 0.25, 0.25));

	  if((center.dot(n) + d) <= 0)
	    ;	      
	  else nabove++;
	}
	  
	// Only split when it changes the mesh.
	if (nabove != 0 && nabove != (int)neighs.size()) {
	  printf("fracture: %d\n", ti);	

	  // Add the new vertices.
	  vert.push_back(vert[vi]);
	  restVert.push_back(restVert[vi]);
	  vs.push_back(vs[vi]);
	
	  vector<int> above;
	  vector<int> below;

	  for (int i=0; i<(int)neighs.size(); i++) {
	    Vec3f center = tets[neighs[i]].getWorldCoordinates(restVert, Vec4f(0.25, 0.25, 0.25, 0.25));
	      
	    if((center.dot(n) + d) <= 0) {
	      tets[neighs[i]].replaceVert(vi, vert.size() - 1);
	      tets[neighs[i]].color = Vec3f(0, 0, 1);
	      below.push_back(neighs[i]);
	    }
	    else {
	      tets[neighs[i]].color = Vec3f(0, 1, 0);
	      above.push_back(neighs[i]);
	    }

	    // Only  let one fracture take place per-tet per-time-instant 
	    affected.insert(neighs[i]);
	  }

	  // Keep track of the fractures.
	  fractures.push_back(TetMeshFractureInfo(tets, ti, n, d, vi, vert.size() - 1, above, below));

#ifdef DEBUG_FRACTURE
	  fracThresh *= 10000;
#endif
	}
	tets[ti].color += Vec3f(0.2, 0.2, 0.2);
      }
    }
  }
}

const std::vector<TetMeshFractureInfo> & TetMesh::getRecentFractures() const {
  return fractures;
}

void TetMesh::debugFracture(TetMeshFractureInfo & info) {
  fractures.clear();
  fractures.push_back(info);

  int vi = info.vold;

  vert.push_back(vert[vi]);
  restVert.push_back(restVert[vi]);
  vs.push_back(vs[vi]);

  for (int i=0; i<(int)info.tetBelow.size(); i++) {
    tets[info.tetBelow[i]].replaceVert(vi, info.vnew);
  }

  // Don't need to propagate results (only carea about output)
  info.initSeparatingTriangles(tets);
}

    
std::vector<int> TetMesh::getTetrahedronContaining(int vi) {
  std::vector<int> neigh;
    
  for(int i=0; i<(int)tets.size(); i++){
    if(tets[i].containsVertex(vi))
      neigh.push_back(i);
  }
  return neigh;
}


void TetMesh::updateNormals(){
  for(int i=0; i<(int)tets.size(); i++)
    tets[i].updateNormals(vert);
}

std::vector<Matrix> TetMesh::getTetRotations(){
  std::vector<Matrix> Rs;

  for(int ti=0; ti<(int)tets.size(); ti++){
    nacb::Matrix P = tets[ti].getBaryMatrix(restVert);
    nacb::Matrix Q = tets[ti].getBaryMatrix(vert);
      
    nacb::Matrix A = Q*P.inverse();
    nacb::Matrix U, S, V;
      
    A.submatrix(0, 0, 3, 3).Lsvd(U, S, V);
    Matrix R = U * V.transpose();

    A.setSubmatrix(0, 0, R);

    Matrix Rfull = Matrix::zeros(12, 12);
    for(int i=0; i<4; i++)
      Rfull.setSubmatrix(3*i, 3*i, R);
      
    Rs.push_back(Rfull);
  }
    
  return Rs;
}

Vec4f TetMesh::getRestBaryCoordinates(int tet, const Vec3f & point) const {
  Matrix M = tets[tet].getBaryMatrix(restVert);

  nacb::Matrix x(4, 1);
  x[0] = point.x;
  x[1] = point.y;
  x[2] = point.z;
  x[3] = 1.0;
    
  nacb::Matrix co = Matrix::linLeastSq(M, x);
  return (Vec4f)co;
}

Vec4f TetMesh::getBaryCoordinates(int tet, const Vec3f & point) const {
  Matrix M = tets[tet].getBaryMatrix(vert);

  nacb::Matrix x(4, 1);
  x[0] = point.x;
  x[1] = point.y;
  x[2] = point.z;
  x[3] = 1.0;
    
  nacb::Matrix co = Matrix::linLeastSq(M, x);
  return (Vec4f)co;
}

Vec3f TetMesh::getWorldCoordinates(int tet, const Vec4f & bary) const {
  return tets[tet].getWorldCoordinates(vert, bary);
}


int TetMesh::getClosestTetrahedron(const Vec3f & pt, bool show) {
  int best = 0;
  double bestDist = 1e10;

  for (int i=0; i<(int)tets.size(); i++) {
    double dist = tets[i].distance(vert, pt);

    if (show) {
      printf("%d %f (%f)\n", i, dist, bestDist);
    }

    if (dist < bestDist) {
      bestDist = dist;
      best = i;
    }
  }
  printf("bestDist: %f\n", bestDist);
  return best;
}


void TetMesh::simulate(double dt, Matrix fext){
  if (fext.m != vert.size() * 3)
    fext = Matrix::zeros(vert.size()*3, 1);

  std::vector<Matrix> Rs;
  Rs = getTetRotations();
    
  Matrix M = getMassMatrix();
  Matrix C = getDampingMatrix();

  Matrix f0 = Matrix::zeros(vert.size()*3, 1);
  //Matrix fext = Matrix::zeros(vert.size()*3, 1);
  Matrix fplastic = Matrix::zeros(vert.size()*3, 1);
    
  Matrix xi = getCombinedVertVector();
  Matrix vi = getCombinedVelocityVector();
        
        
#ifdef USE_SPARSE
  // Get the sparse matrix (bare-bones entries only need to be made once)
  std::vector<std::vector<std::pair<int, double> > > columns(vert.size()*3);
  //std::vector<std::map<int, int> > rowMap(vert.size()*3);

  std::vector<std::set<int> > vadj = getVertexAdjacency();

  for(int i=0; i<vert.size(); i++){
    std::set<int>::iterator it = vadj[i].begin();
    std::set<int>::iterator last = vadj[i].end();
      
    std::vector<std::pair<int, double> > column;
   
    for(; it != last; it++){
      column.push_back(std::pair<int, double>(*it*3 + 0, 0.0));
      column.push_back(std::pair<int, double>(*it*3 + 1, 0.0));
      column.push_back(std::pair<int, double>(*it*3 + 2, 0.0));
    }

    std::sort(column.begin(), column.end());
      
    for(int k=0; k<3; k++)
      columns[3*i + k] = column;
      
    /*
      for(int j=0; j<column.size(); j++){
      rowMap[3*i + 0][column[j].first] = j;
      rowMap[3*i + 1][column[j].first] = j;
      rowMap[3*i + 2][column[j].first] = j;
      }
    */
    // Later: 
    //columns[3*vj + jj][rowMap[3*vj + jj][3*vi + ii]].second += RKs(i, j);

  }
    
  SparseMatrix K(3*vert.size(), columns);
#else
  Matrix K = Matrix::zeros(vert.size()*3, vert.size()*3);
#endif
  nacb::StopWatch timer;

  for(int ti=0; ti<tets.size(); ti++){
      
    Matrix R = Rs[ti];
    Matrix Rinv = R.transpose();
    Matrix RKs = R * tets[ti].K * Rinv;
    Matrix RKx0 = R * (tets[ti].K * tets[ti].getVertexVector(restVert));
      
    Matrix RPe_plastic;
    bool have_plastic = true;
      
    if (have_plastic) {
      tets[ti].updatePlasticity(restVert, vert, Rinv, dt);
	
      RPe_plastic = R * tets[ti].P * tets[ti].e_plastic;
    }

    //RKs.printMatlab("K");
    //RKx0.printMatlab("RKx0");

    for(int i=0; i<12; i++){
      int vi = tets[ti].vi[i/3];
      int ii = i%3;

      for(int j=0; j<12; j++){
	int vj = tets[ti].vi[j/3];
	int jj = j%3;
	
	K(3*vi + ii, 3*vj + jj) += RKs(i, j);
      }

      f0[3*vi + ii] -= RKx0[i];
	
      if (have_plastic)
	fplastic[3*vi + ii] += RPe_plastic[i];
    }
  }
  typeof(K) A = K*dt*dt;

  for(int i=0; i<vert.size()*3; i++)
    A(i, i) += M.data[i] + C.data[i]*dt;

  Matrix b = M;
  M.dotmult(vi);
  b -= (K*xi + f0 - fplastic - fext)*dt;

#ifdef USE_SPARSE
  Matrix vnew = SparseMatrix::umfsolve(A, b);
#else
  Matrix vnew = Matrix::LlinLeastSq(A, b);
#endif

  // M and C used to be matrices, now are vectors for convenience.
  // Matrix A = M + C*dt + K*dt*dt;
  // Matrix b = M*vi - (K*xi + f0 + fplastic - fext)*dt;
  // Matrix vnew = Matrix::LlinLeastSq(A, b);

  Vec3f vavg(0, 0, 0);
  for(int i=0; i<vert.size(); i++){
    vs[i] = Vec3f(vnew[3*i+0], vnew[3*i+1], vnew[3*i+2]);
    vert[i] += vs[i]*dt;

    vavg += vs[i];
  }
  vavg *= (1.0/vert.size());

  handleFractures(Rs);
    
  // Quick collision hack...remove this
  for(int i=0; i<vert.size(); i++){
    if(vert[i].y < kGroundHeight && vs[i].y < 0) {
      vs[i].y = -vs[i].y;
      vs[i] *= 0.9; // Friction  :)
    }
  }
  updateNormals();
}
