#ifndef TET_MESH_H
#define TET_MESH_H

#include <GL/glew.h>

#include "UI.h"
#include "Globals.h"
#include "Triangle.h"
#include "TriangleIndices.h"

#include <ik/mesh.h>
#include <nmath/matrix.h>

#include <vector>
#include <set>
#include <stdlib.h>
#include <GL/gl.h>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/set.hpp>
#include <boost/shared_ptr.hpp>

#include "Tetrahedron.h"
#include "TetMeshFractureInfo.h"


using namespace nacb;
using namespace std;

class TetMesh: public Mesh {

public:
  std::vector<Tetrahedron> tets;
  std::vector<Vec3f> vs;  
  std::vector<TetMeshFractureInfo> fractures;

public:
  double fracThresh;
  Tetrahedron::Material defaultMaterial;

  TetMesh(const Tetrahedron::Material & mat = Tetrahedron::Material());
  
  int loadobj(const char * fname);
  void removeDuplicateVert();

  // Assume that file format could not represent the 
  // tetrahedron (so merge any vertices that are connected).
  void buildTetrahedron();
  
  void draw();
  void drawSelect();
  Matrix getMassMatrix() const;
  Matrix getDampingMatrix() const;

  Matrix getCombinedVertVector() const;
  Matrix getCombinedVelocityVector() const;

  std::vector<std::set<int> > getVertexAdjacency();

  void simulate(double dt, Matrix fext = Matrix());

  void handleFractures(const vector<Matrix> & Rs);
  const std::vector<TetMeshFractureInfo> & getRecentFractures() const;
  std::vector<int> getTetrahedronContaining(int vi);

  void updateNormals();
  std::vector<Matrix> getTetRotations();

  Vec4f getRestBaryCoordinates(int tet, const Vec3f & point) const;
  Vec4f getBaryCoordinates(int tet, const Vec3f & point) const;
  Vec3f getWorldCoordinates(int tet, const Vec4f & bary) const;
  int getClosestTetrahedron(const Vec3f & pt, bool show = false);


  const Tetrahedron::Material getMaterial() const {
    return defaultMaterial;
  }

  void setMaterial(Tetrahedron::Material & mat) {
    defaultMaterial = mat;
    updateTetVolume();
  }

  void resetToRest(){
    for (int i=0; i<vs.size(); i++) {
      vs[i] = Vec3f(0, 0, 0);
      tets[i].e_plastic *= 0;
    }
    vert = restVert;
  }

  // Used to simulate the fractures (from a file, to play back fractures).
  void debugFracture(TetMeshFractureInfo & info);

private:  
  
  void updateTetVolume(){
    for(int i=0; i<(int)tets.size(); i++){
      tets[i].plasticity = defaultMaterial;
      tets[i].updateVolume(restVert);
    }
  }

  bool anyTetExists(std::set<int> & tetExists, int v0, int v1, int v2, int v3){
    int perms[24][4] = {{0, 1, 2, 3}, {1, 0, 2, 3}, {0, 2, 1, 3}, {1, 2, 0, 3},
			{2, 0, 1, 3}, {2, 1, 0, 3}, {0, 1, 3, 2}, {1, 0, 3, 2},
			{0, 2, 3, 1}, {1, 2, 3, 0}, {2, 0, 3, 1}, {2, 1, 3, 0},
			{0, 3, 1, 2}, {1, 3, 0, 2}, {0, 3, 2, 1}, {1, 3, 2, 0},
			{2, 3, 0, 1}, {2, 3, 1, 0}, {3, 0, 1, 2}, {3, 1, 0, 2},
			{3, 0, 2, 1}, {3, 1, 2, 0}, {3, 2, 0, 1}, {3, 2, 1, 0}};

    int vi[4] = {v0, v1, v2, v3};
    
    for(int k=0; k<24; k++){
      if(tetExists.count(tetHash(vi[perms[k][0]],
				 vi[perms[k][1]],
				 vi[perms[k][2]],
				 vi[perms[k][3]])))
	return true;
    }
    return false;
  }

  int tetHash(int v0, int v1, int v2, int v3){
    return faceHash(v0, v1, v2) + (vert.size()*vert.size()*vert.size())*v3;
  }

  int faceHash(int v0, int v1, int v2){
    return v0*(vert.size()*vert.size())+ v1*vert.size() + v2;
  }

public:
  typedef boost::shared_ptr<TetMesh> ptr;
};


#endif //TET_MESH_H
