#include "TetMeshFractureInfo.h"
#include "TetMesh.h"

TetMeshFractureInfo::TetMeshFractureInfo(const std::vector<Tetrahedron> & tets,
					 const int _ti,
					 const Vec3f & _n,
					 const double _d,
					 int _vold, int _vnew,
					 const std::vector<int> & _tetAbove,
					 const std::vector<int> & _tetBelow): 
  ti(_ti), n(_n), d(_d), vold(_vold), vnew(_vnew), 
  tetAbove(_tetAbove), tetBelow(_tetBelow) { 
  initSeparatingTriangles(tets);
  tetAboveSet = std::set<int>(tetAbove.begin(), tetAbove.end());
  tetBelowSet = std::set<int>(tetBelow.begin(), tetBelow.end());
}


bool TetMeshFractureInfo::affectsTetrahedron(int t) const {
  return tetAboveSet.count(t) || tetBelowSet.count(t);
}

    
void TetMeshFractureInfo::initSeparatingTriangles(const std::vector<Tetrahedron> & tets) {
  sepTris.clear();
  sepEdges.clear();

  std::set<pair<int, int> > sepTriEdges;

  std::cout << "Above: ";
  for (int i=0; i<(int)tetAbove.size(); i++) {
    tets[tetAbove[i]].print(std::cout) << "\n"; 
  }

  std::cout << "Below: ";
  for (int i=0; i<(int)tetBelow.size(); i++) {
    tets[tetBelow[i]].print(std::cout) << "\n";
  }

  // FIXME: this could be sped-up
  // find a pair s.t. triangle vold-v1-v2 and vnew-v2-v1 exist
  for (int i=0; i<(int)tetAbove.size(); i++){
    for (int j=0; j<(int)tetBelow.size(); j++){
      int firstTriIndex;
      if (adjacent(tets[tetAbove[i]], tets[tetBelow[j]], firstTriIndex)) {
	sepTris.push_back(SeparatingTriangle(tetAbove[i], tetBelow[j], 
					     firstTriIndex));
	printf("Found sep tri: %d/%d %d/%d\n", 
	       i, (int)tetBelow.size(), j, (int)tetAbove.size());

	TriangleIndices tri = Tetrahedron::tris[firstTriIndex].mapIndices(tets[tetAbove[i]].vi);

	for (int k=0; k<3; k++) {
	  int v1 = tri[k], v2 = tri[(k + 1)%3];
	  sepTriEdges.insert(make_pair(v1, v2));
	  sepTriEdges.insert(make_pair(v2, v1));
	}

	assert(tri.replaceVert(vold, vnew));
	    
	// Do the same for the new vert edges.
	for (int k=0; k<3; k++) {
	  int v1 = tri[k], v2 = tri[(k + 1)%3];
	  sepTriEdges.insert(make_pair(v1, v2));
	  sepTriEdges.insert(make_pair(v2, v1));
	}
      }
    }
  }      

  // Now find any separating edges (e.g., those edges that are not in
  // a separating triangle, but are still separating).
  for (int i=0; i<(int)tetAbove.size(); i++) {
    for (int j=0; j<(int)tetBelow.size(); j++) {
      findSeparatingEdges(tets[tetAbove[i]], tets[tetBelow[j]], sepTriEdges);
    }
  }
  // Print out some debug information.
  if (!sepTris.size() ) {
    /*
      printf("No sep tri.  Above: %d.  Below: %d (%d, %d)\n", tetAbove.size(), tetBelow.size(), vold, vnew);
      printf("above:\n");
      for (int i=0; i<tetAbove.size(); i++) {
      for (int k1=0; k1<4; k1++) {
      TriangleIndices tri = Tetrahedron::tris[k1].mapIndices(tets[tetAbove[i]].vi);
      printf("%d %d %d\n", tri[0], tri[1], tri[2]);
      }
      }
      printf("below:\n");
      for (int i=0; i<tetBelow.size(); i++) {
      for (int k1=0; k1<4; k1++) {
      TriangleIndices tri = Tetrahedron::tris[k1].mapIndices(tets[tetBelow[i]].vi);
      tri.replaceVert(vnew, vold);
      printf("%d %d %d\n", tri[0], tri[1], tri[2]);
      }
      }
    */
  }
  else {
    printf("Sep tri:\n");
    for (int i=0; i<(int)sepTris.size(); i++) {
      TriangleIndices tri = 
	Tetrahedron::tris[sepTris[i].tetAboveTriIndex].mapIndices(tets[sepTris[i].tetAbove].vi);
      printf("%d %d %d  (%d, %d)\n", tri[0], tri[1], tri[2], vold, vnew);
    }
  }
}  


void TetMeshFractureInfo::findSeparatingEdges(const Tetrahedron & t1,
					      const Tetrahedron & t2,
					      std::set<pair<int, int> > & exclude) {
  int v1ind = t1.indexOfVertex(vold);
  int v2ind = t2.indexOfVertex(vnew);

  assert(v1ind >=0 && v2ind >= 0);

  for (int i1=0; i1<4; i1++) {
    if (i1 == v1ind) continue;
	
    for (int i2=0; i2<4; i2++) {
      if (i2 == v2ind) continue;
	  
      if (t1.vi[i1] == t2.vi[i2] && !exclude.count(make_pair(vnew, t1.vi[i1]))) {
	printf("Sep edge: %d-%d %d-%d\n", vold, t1.vi[i1], vnew, t2.vi[i2]);
	printf("  %d %d %d %d  (%d => %d, %d) \n", t1.vi[0], t1.vi[1], t1.vi[2], t1.vi[3],
	       v1ind, t1.vi[v1ind], vold);
	    
	sepEdges.push_back(SeparatingEdge(vold, vnew, t1.vi[i1]));
	exclude.insert(make_pair(vnew, t1.vi[i1]));
	exclude.insert(make_pair(t1.vi[i1], vnew));

	exclude.insert(make_pair(vold, t1.vi[i1]));
	exclude.insert(make_pair(t1.vi[i1], vold));
      }
    }
  }
}


bool TetMeshFractureInfo::adjacent(const Tetrahedron & t1,
				   const Tetrahedron & t2, int & match) {
  // See if the tetrahedron are adjacent (need to make sure that we take
  // the vold and vnew into account.
  for (int k1=0; k1<4; k1++) {
    TriangleIndices tri1 = Tetrahedron::tris[k1].mapIndices(t1.vi);
    if(!tri1.makeFirst(vold))
      continue;
	
    for (int k2=0; k2<4; k2++) {
      TriangleIndices tri2 = Tetrahedron::tris[k2].mapIndices(t2.vi);
      if (opposingTriangles(tri1, tri2)) {
	match = k1;
	return true;
      }
    }
  }
  return false;
}


// Assumes tri1[0] = vold
bool TetMeshFractureInfo::opposingTriangles(const TriangleIndices & tri1, const TriangleIndices & tri2){
  for (int k=0; k<3; k++) {
    if (vnew == tri2[k]) {
      return (tri1[2] == tri2[(k + 1)%3]) && (tri1[1] == tri2[(k + 2)%3]);
    }
  }
  return false;
}
