#ifndef FRACTURE_INFO_H
#define FRACTURE_INFO_H

#include "TriangleIndices.h"

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/set.hpp>
#include <boost/shared_ptr.hpp>

#include <vector>
#include <set>
#include <nmath/vec3.h>

using namespace nacb;
using namespace std;

class Tetrahedron;

/**
   \brief Structure for maintaining tetrahedra that
   have recently been split (so that bound meshes can
   subsequently process their split information).
*/
class TetMeshFractureInfo {
 public:
  int   ti;
  Vec3f  n;
  double d;
  int vold;
  int vnew;
  std::vector<int> tetAbove;
  std::vector<int> tetBelow;

  std::set<int> tetAboveSet;
  std::set<int> tetBelowSet;

  struct SeparatingTriangle {
    int tetAbove;
    int tetBelow;
    
    int tetAboveTriIndex;

    SeparatingTriangle(int _tetAbove = -1,
		       int _tetBelow = -1,
		       int _tetAboveTriIndex = -1) :
    tetAbove(_tetAbove), tetBelow(_tetBelow), tetAboveTriIndex(_tetAboveTriIndex) {  }

    friend class boost::serialization::access;
      
    template<class Archive>
    void serialize(Archive & ar, const unsigned int version) {
      ar & tetAbove & tetBelow & tetAboveTriIndex;
    }      
  };

  struct SeparatingEdge {
    int vold, vnew;
    int vother;

    SeparatingEdge(int _vold = -1, int _vnew = -1, int _vother = -1) : 
     vold(_vold), vnew(_vnew), vother(_vother) {

    }

    bool onEdge(const vector<Vec3f> & vert, const Vec3f & pt, double eps = 1e-5) const {
      Vec3f diff = pt - vert[vold];
      Vec3f edge = vert[vother] - vert[vold];
	
      edge.normalize();
      double t = diff.dot(edge);
	
      diff -= edge*t;
      return diff.len() < eps;
    }
  };

  vector<SeparatingTriangle> sepTris;
  vector<SeparatingEdge>    sepEdges;

  TetMeshFractureInfo() { }

  TetMeshFractureInfo(const std::vector<Tetrahedron> & tets,
		      const int _ti,
		      const Vec3f & _n,
		      const double _d,
		      int _vold, int _vnew,
		      const std::vector<int> & _tetAbove,
		      const std::vector<int> & _tetBelow);

  bool affectsTetrahedron(int t) const;
    
  void initSeparatingTriangles(const std::vector<Tetrahedron> & tets);

  void findSeparatingEdges(const Tetrahedron & t1,
			   const Tetrahedron & t2,
			   std::set<pair<int, int> > & exclude);

  bool adjacent(const Tetrahedron & t1,
		const Tetrahedron & t2, int & match);

  // Assumes tri1[0] = vold
  bool opposingTriangles(const TriangleIndices & tri1, const TriangleIndices & tri2);

 private:
  friend class boost::serialization::access;
    
  template<class Archive>
    void serialize(Archive & ar, const unsigned int version) {
    ar & ti;
    ar & n.x & n.y & n.z & d;
    ar & vold & vnew;
    ar & tetAbove & tetBelow & tetAboveSet & tetBelowSet;
    ar & sepTris;
  }    
  //BOOST_SERIALIZATION_SPLIT_MEMBER();
};


#endif
