#include "Tetrahedron.h"
#include "Triangle.h"
#include "UI.h"
#include <GL/gl.h>

using namespace nacb;
using namespace std;

const TriangleIndices Tetrahedron::tris[4] = {TriangleIndices(0, 1, 2), 
						       TriangleIndices(1, 0, 3), 
						       TriangleIndices(2, 1, 3), 
						       TriangleIndices(0, 2, 3)};


Tetrahedron::Tetrahedron() : K(Matrix::eye(6, 12)),  
				      B(Matrix::zeros(6, 12)), 
				      E(Matrix::zeros(6, 6)),
				      P(Matrix::zeros(12, 6)), 
				      e_plastic(Matrix::zeros(6, 1)) {  }

    
Tetrahedron::Tetrahedron(int v1, int v2, int v3, int v4){ 
  vi[0] = v1;
  vi[1] = v3 ;
  vi[2] = v4;
  vi[3] = v2;

  color = Vec3f(0.7, 0.7, 0.7);//drand48()*0.5 + 0.5, drand48(), drand48());
  V = 0;
      
  B = Matrix::zeros(6, 12);
  K = Matrix::eye(6, 12);
  P = Matrix::zeros(12, 6);
  E = Matrix::zeros(6, 6);

  e_plastic = Matrix::zeros(6, 1);
}


bool Tetrahedron::replaceVert(int vold, int vnew){
  for (int i=0; i<4; i++) {
    if(vi[i] == vold) {
      vi[i] = vnew;
      return true;
    }
  }
  return false;
}


bool Tetrahedron::containsVertex(int v) const {
  return indexOfVertex(v) >= 0;
}


int Tetrahedron::indexOfVertex(int v) const {
  for (int i=0; i<4; i++) {
    if(vi[i] == v)
      return i;
  }
  return -1;
}


void Tetrahedron::updatePlasticity(const std::vector<Vec3f> & restVert,
					    const std::vector<Vec3f> & vert,
					    const Matrix & Rinv,
					    double dt) {
  Matrix e_total = B*(Rinv*getVertexVector(vert) - getVertexVector(restVert));
  Matrix e_elastic = e_total - e_plastic;
      
  if(sqrt(e_elastic.dot(e_elastic)) > plasticity.c_yield) {
    e_plastic +=  e_elastic * (plasticity.c_creep * dt);
  }
      
  double e_norm = sqrt(e_plastic.dot(e_plastic));
  if(e_norm > plasticity.c_max)
    e_plastic *= (plasticity.c_max / e_norm);
}


void Tetrahedron::updateNormals(const std::vector<Vec3f> & vert){
  for(int k=0; k<4; k++){
    Vec3f e1 = vert[vi[tris[k][1]]] - vert[vi[tris[k][0]]];
    Vec3f e2 = vert[vi[tris[k][2]]] - vert[vi[tris[k][1]]];
	
    normals[k] = (e1.cross(e2));
    normals[k].normalize();
  }
}


void Tetrahedron::drawSelect(const std::vector<Vec3f> & vert){
      
  glBegin(GL_TRIANGLES);
  for(int k=0; k<4; k++) {
    glVertex3fv(vert[vi[Tetrahedron::tris[k][0]]].data);
    glVertex3fv(vert[vi[Tetrahedron::tris[k][1]]].data);
    glVertex3fv(vert[vi[Tetrahedron::tris[k][2]]].data);
  }	
  glEnd();
}


void Tetrahedron::draw(const std::vector<Vec3f> & vert){
  glColor3fv(color.data);
      
  glBegin(GL_TRIANGLES);
  for(int k=0; k<4; k++) {
    glNormal3fv(normals[k].data);

    glVertex3fv(vert[vi[Tetrahedron::tris[k][0]]].data);
    glVertex3fv(vert[vi[Tetrahedron::tris[k][1]]].data);
    glVertex3fv(vert[vi[Tetrahedron::tris[k][2]]].data);
  }	
  glEnd();


  glColor3f(1, 1, 1);
  for (int i=0; i<(int)vert.size(); i++) {
    glPushMatrix();
    glTranslatef(vert[i].x, vert[i].y, vert[i].z);
    UI::drawString(0, 0, "%d", i);
    glPopMatrix();
  }
}


bool Tetrahedron::intersectSegment(const std::vector<Vec3f> & vert, 
				   const Vec3f & p1, const Vec3f & p2, 
				   double & t, double eps) const {
  t = 1;
  bool succ = false;
  for (int k=0; k<4; k++) {
    Triangle tri(vert, Tetrahedron::tris[k].mapIndices(vi));
    double tuse;
    if (tri.intersectSegment(p1, p2, tuse, eps) ){
      t = std::min(t, tuse);
      succ = true;
    }
  }
  return succ;
}


Matrix Tetrahedron::getVertexVector(const std::vector<Vec3f> & vert) const {
  Matrix x(12, 1);
      
  for(int i=0; i<4; i++){
    for(int j=0; j<3; j++)
      x[3*i + j] = vert[vi[i]].data[j];
  }
  return x;
}


Vec3f Tetrahedron::getWorldCoordinates(const std::vector<Vec3f> & vert,
				       const Vec4f & bary) const {
  Vec3f co(0, 0, 0);
      
  for (int j=0; j<4; j++)
    co += vert[vi[j]] * bary.data[j];
      
  return co;
}


double Tetrahedron::distanceToTriangle(const std::vector<Vec3f> & vert,
				       int tri, const Vec3f & pt) const {
  double dist = 1e10;

  for (int j=0; j<3; j++) {
    Vec3f diff =  pt - vert[tris[tri][j]];
    dist = std::min(dist, diff.len());
  }

  if (dist <= 1e-8) {
    printf("triangle points coincide.\n");
  }
      
  nacb::Matrix M(3, 3);
  Vec3f v1 = vert[vi[tris[tri][0]]];
  Vec3f v2 = vert[vi[tris[tri][1]]];
  Vec3f v3 = vert[vi[tris[tri][2]]];

  M(0, 0) = v1.x;
  M(1, 0) = v1.y;
  M(2, 0) = v1.z;

  M(0, 1) = v2.x;
  M(1, 1) = v2.y;
  M(2, 1) = v2.z;
      
  M(0, 2) = v3.x;
  M(1, 2) = v3.y;
  M(2, 2) = v3.z;

  nacb::Matrix b(3, 1);

  b[0] = pt.x;
  b[1] = pt.y;
  b[2] = pt.z;

  // Use the minimum clamped point as the distance,
  // this is not entirely correct (should use the lines).
  nacb::Matrix x = nacb::Matrix::linLeastSq(M, b);

  x[0] = std::max(0., std::min(x[0], 1.0));
  x[1] = std::max(0., std::min(x[1], 1.0));
  x[2] = std::max(0., std::min(x[2], 1.0));
      

  double sum = x[0] + x[1] + x[2];
  x *= (1.0/sum);


  Vec3f tript = (v1*x[0] + v2*x[1] + v3*x[2]);
  dist = std::min((tript - pt).len(), dist);

  if (dist <= 1e-7) {
    std::cout << v1 << v2 << v3 << "\n";
    printf("Triangle distance zero: %f %f %f   %f %f %f\n", x[0], x[1], x[2], 
	   x[1], x[2], 1.0 - x[1] - x[2]);
    std::cout << tript << "\n";
    std::cout << pt << "\n";
  }

  return dist;
}

    
double Tetrahedron::distance(const std::vector<Vec3f> & vert,
				      const Vec3f & pt) const {
  double dist = 1e10;

  for (int j=0; j<4; j++) {
    double distTri = distanceToTriangle(vert, j, pt);
    if (!isnan(distTri) && distTri < dist)
      dist = distTri;
  }
      
  nacb::Matrix M = getBaryMatrix(vert);
  nacb::Matrix x(4, 1);
  x[0] = pt.x;
  x[1] = pt.y;
  x[2] = pt.z;
  x[3] = 1.0;

  nacb::Matrix co = Matrix::linLeastSq(M, x);
      
  double sum = 0;
  for (int k=0; k<4; k++) {
    co[k] = std::max(std::min(co[k], 1.0), 0.0);
    sum += co[k];
  }
  sum = std::max(sum, 1e-6);
  co *= (1.0/sum);
      
  nacb::Matrix res = M * co - x;
  res[3] = 0;

  double distTet = sqrt(res.dot(res));
      
  if (distTet < dist) 
    dist = distTet;

  return dist;
}


Matrix Tetrahedron::getBaryMatrix(const std::vector<Vec3f> & vert) const {
  Matrix M(4, 4);

  for(int j=0; j<4; j++){
    M(0, j) = vert[vi[j]].x;
    M(1, j) = vert[vi[j]].y;
    M(2, j) = vert[vi[j]].z;
    M(3, j) = 1;
  }
      
  return M;
}


void Tetrahedron::updateVolume(const std::vector<Vec3f> & vert){
  Vec3f a = vert[vi[1]] - vert[vi[0]];
  Vec3f b = vert[vi[2]] - vert[vi[0]];
  Vec3f c = vert[vi[3]] - vert[vi[0]];

  V  = -(a.dot(b.cross(c)))/6.0;
      
  /*
    The stiffness matrix (or strain) is bunch of displacements derivatives w.r.t
    the x, y, and z coords.  As the displacements are modeled over the element 
    linearly (e.g., barycentric weighted over the element), all these derivatives
    are proportional to the barycentric coordinates of a point in the tet.  The
    bary centric coordinates are obtained by inverting the edge matrix.  These
    coefficients are plugged into the strain tensor (the displacements at
    the 0-th coordinate is kappa = (1 - alpha*x - gamma*x - beta*x)
    meaning the derivative contains -alpha - gamma - beta, hence the computation below.
	
    The other stresses (e.g.,gamma_{x,z}) (just re-arrange the values ones).

    The book didn't give a derivation, but I think this is correct.
    Really want derivatives in terms of the (x,y,z) coordinates of the node.
  */
  {
    double a1 = a.x, a2 = b.x, a3 = c.x;
    double b1 = a.y, b2 = b.y, b3 = c.y;
    double c1 = a.z, c2 = b.z, c3 = c.z;
	
    B = Matrix::zeros(6, 12);
    double det = (a1*b2*c3-a1*b3*c2-b1*a2*c3+b1*a3*c2+c1*a2*b3-c1*a3*b2);
	
    // alpha = [b2*c3-b3*c2, -a2*c3+a3*c2,  a2*b3-a3*b2]/det * [x,y,z]
    // beta = [ -b1*c3+b3*c1,  a1*c3-a3*c1,  b1*a3-a1*b3]/det * [x,y,z]
    // gamma = [  b1*c2-b2*c1, -a1*c2+a2*c1,  a1*b2-a2*b1]/det * [x,y,z]
	
    B(0, 3) =  (b2*c3-b3*c2)/det;
    B(0, 6) = (-b1*c3+b3*c1)/det;
    B(0, 9) =  (b1*c2-b2*c1)/det;
    B(0, 0) = -B(0, 3) - B(0, 6) - B(0, 9);
	
	
    B(1, 4) =  (-a2*c3+a3*c2)/det;
    B(1, 7) =  (a1*c3-a3*c1)/det;
    B(1, 10) =  (-a1*c2+a2*c1)/det;
    B(1, 1) = -B(1, 4) - B(1, 7) - B(1, 10);
	
	
    B(2, 5) =  (a2*b3-a3*b2)/det;
    B(2, 8) =  (b1*a3-a1*b3)/det;
    B(2, 11) =  (a1*b2-a2*b1)/det;
    B(2, 2) = -B(2, 5) - B(2, 8) - B(2, 11);
	
	
    // gamma_{x,z} 
    for(int k=0; k<4; k++){
      B(3, 3*k + 0) =  B(2, 3*k + 2);
      B(3, 3*k + 2) =  B(0, 3*k + 0);
    }


    // gamma_{x,y} 
    for(int k=0; k<4; k++){
      B(4, 3*k + 0) =  B(1, 3*k + 1);
      B(4, 3*k + 1) =  B(0, 3*k + 0);
    }
	
	
    // gamma_{y,z} 
    for(int k=0; k<4; k++){
      B(5, 3*k + 1) =  B(2, 3*k + 2);
      B(5, 3*k + 2) =  B(1, 3*k + 1);
    }
    
    // B.printMatlab("B");

    E = Matrix::zeros(6, 6);
    double e = plasticity.e;//25000; //250000// Elastic modulus ???
    double v = plasticity.v;//0.1;//0.43; // Poisson's ratio
    double c = e/((1 + v)*(1 - 2.*v));
    double G = e/(2.*(1 + v));

    E(0, 0) = E(1, 1) = E(2, 2) = (1.0-v)*c;
    E(0, 1) = E(0, 2) = E(1, 2) = v*c;
    E(1, 0) = E(2, 0) = E(2, 1) = v*c;
    E(3, 3) = E(4, 4) = E(5, 5) = G;

    K = B.transpose()*E*B*V;
	
    //E.printMatlab("E");
    //K.printMatlab("K");
    // Update the plasticity
    P = ((B.transpose()) * E)*V;
  }
  // printf("V: %lf\n", V);      
}


std::ostream & Tetrahedron::print(std::ostream & o) const {
  o << vi[0] << "," << vi[1] << "," << vi[2] << "," << vi[3];
  return o;
}
  

