#ifndef TRIANGLE_H
#define TRIANGLE_H

#include "TriangleIndices.h"

#include <nmath/vec3.h>
#include <nmath/vec2.h>

#include <vector>
#include <stdlib.h>
#include <string.h>

using namespace nacb;
using namespace std;

/**
   Triangle helper class.  Useful for intersecting rays and segments
   with tolerances on the intersections.
 */
class Triangle {
public:
  Vec3f vert[3]; //!< The vertices.
  Vec3f e1, e2; //!< Edges between vert[0]-vert[1] and vert[0]-vert[2]
  Vec3f d1, d2; //!< Temp variables used for projecting into triangle
  Vec3f n;  //!< Triangle normal.
  double d; //!< Plane equation n'*x + d = 0

  Triangle() {
    memset(this, 0, sizeof(Triangle));
  }

  /**
     Create a new triangle by indexing the vertex list.
  */
  Triangle(const vector<nacb::Vec3f> & verts, const TriangleIndices & inds){
    vert[0] = verts[inds[0]];
    vert[1] = verts[inds[1]];
    vert[2] = verts[inds[2]];
    e1 = vert[1] - vert[0];
    e2 = vert[2] - vert[0];

    n = e1.cross(e2);
    n.normalize();

    d = -vert[0].dot(n);

    
    double A, B, C, det;

    A = e1.dot(e1);
    B = e1.dot(e2);
    C = e2.dot(e2);
    det = A*C - B*B;
    
    d1 = (e1*C - e2*B)*(1.0/det);
    d2 = (e1*(-B) + e2*A)*(1.0/det);
  }
  
  /**
     \return true if the point is below the triangle.
  */
  bool below(const Vec3f & pt) const {
    return n.dot(pt) + d < 0;
  }

  /**
     \return true if the point is above the triangle.
   */
  bool above(const Vec3f & pt) const {
    return !below(pt);
  }

  /**
     \return the index of the edge that the vertex is on (or -1 if it isn't)
  */
  int onEdge(const Vec3f & co, double eps = 1e-5) {
    for (int k=0; k<3; k++) {
      Vec3f diff = co - vert[k];
      
      Vec3f edge =  vert[(k + 1) % 3] - vert[k];
      edge.normalize();
      
      double t = diff.dot(edge);
      double dist = (diff - edge*t).len();

      if (dist < eps)
	return k;
    }
    return -1;
  }

  /**
     Intersect the segment v1-v2 with the triangle. 
     \param t output parameter
     \return true if the segment intersects the triangle.
             On output the linear combination parameter, t, will
	     be set appropriately.
  */
  bool intersectSegment(const Vec3f & v1, const Vec3f & v2, double & t, 
			const double eps = -1e-8) const {
    if (intersectRay(v1, v2 - v1, t)) {
      return (t >= -eps && t <= 1 + eps);
    }    
    return false;
  }
  
  /**
     Intersect the ray v1 + t*k with the triangle. 
     \return true if the ray intersects the triangle.
             On output the linear combination parameter, t, will
	     be set appropriately.
  */
  bool intersectRay(const Vec3f & v1, const Vec3f & k, double & t,
		    const double eps = -1e-8) const {
    t = (-d - n.dot(v1))/n.dot(k);

    Vec3f pt = v1 + k*t;
    pt -= vert[0];
    
    Vec2f p(d1.dot(pt), d2.dot(pt));
    return (p.x >= -eps && p.y >= -eps && (p.x + p.y <= 1+eps));
  }

  /**
     \param  p the point to compute the distance from.
     \return  the distance from the point to the closest point on the triangle.
   */
  double distance(const Vec3f & p) const {
    Vec3f pon = p - vert[0];
    pon -= n*(n.dot(pon));

    double a = d1.dot(pon), b = d2.dot(pon);
    a = std::max(0.0, a);
    b = std::max(0.0, b);

    double sum = a + b;
    if (sum >= 1.0) {
      a /= sum;
      b /= sum;
    }
    Vec3f diff = (vert[0] + e1*a + e2*b) - p;
    return diff.len();
  }
};


#endif // TRIANGLE_H
